"""Admin routes for presentations."""

import json

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate
from marshmallow.validate import Range

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....indy.holder import IndyHolder, IndyHolderError
from ....indy.models.cred_precis import IndyCredPrecisSchema
from ....indy.models.pres_preview import IndyPresPreview, IndyPresPreviewSchema
from ....indy.models.proof import IndyPresSpecSchema
from ....indy.models.proof_request import IndyProofRequestSchema
from ....indy.util import generate_pr_nonce
from ....ledger.error import LedgerError
from ....messaging.decorators.attach_decorator import AttachDecorator
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.models.paginated_query import (
    PaginatedQuerySchema,
    get_paginated_query_params,
)
from ....messaging.valid import (
    INDY_EXTRA_WQL_EXAMPLE,
    INDY_EXTRA_WQL_VALIDATE,
    NUM_STR_NATURAL_EXAMPLE,
    NUM_STR_NATURAL_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
)
from ....storage.base import DEFAULT_PAGE_SIZE, MAXIMUM_PAGE_SIZE
from ....storage.error import StorageError, StorageNotFoundError
from ....utils.tracing import AdminAPIMessageTracingSchema, get_timer, trace_event
from ....wallet.error import WalletNotFoundError
from . import problem_report_for_record, report_problem
from .manager import PresentationManager, PresentationManagerError
from .message_types import ATTACH_DECO_IDS, PRESENTATION_REQUEST, SPEC_URI
from .messages.presentation_problem_report import ProblemReportReason
from .messages.presentation_proposal import PresentationProposal
from .messages.presentation_request import PresentationRequest
from .models.presentation_exchange import (
    V10PresentationExchange,
    V10PresentationExchangeSchema,
)


class V10PresentProofModuleResponseSchema(OpenAPISchema):
    """Response schema for Present Proof Module."""


class V10PresentationExchangeListQueryStringSchema(PaginatedQuerySchema):
    """Parameters and validators for presentation exchange list query."""

    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread identifier", "example": UUID4_EXAMPLE},
    )
    role = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V10PresentationExchange, m)
                for m in vars(V10PresentationExchange)
                if m.startswith("ROLE_")
            ]
        ),
        metadata={"description": "Role assigned in presentation exchange"},
    )
    state = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V10PresentationExchange, m)
                for m in vars(V10PresentationExchange)
                if m.startswith("STATE_")
            ]
        ),
        metadata={"description": "Presentation exchange state"},
    )


class V10PresentationExchangeListSchema(OpenAPISchema):
    """Result schema for an Aries RFC 37 v1.0 presentation exchange query."""

    results = fields.List(
        fields.Nested(V10PresentationExchangeSchema()),
        metadata={"description": "Aries RFC 37 v1.0 presentation exchange records"},
    )


class V10PresentationSendRequestSchema(IndyPresSpecSchema):
    """Request schema for sending a presentation."""

    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )


class V10PresentationProposalRequestSchema(AdminAPIMessageTracingSchema):
    """Request schema for sending a presentation proposal admin message."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    presentation_proposal = fields.Nested(IndyPresPreviewSchema(), required=True)
    auto_present = fields.Boolean(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to respond automatically to presentation requests, building"
                " and presenting requested proof"
            )
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )


class V10PresentationCreateRequestRequestSchema(AdminAPIMessageTracingSchema):
    """Request schema for creating a proof request free of any connection."""

    proof_request = fields.Nested(IndyProofRequestSchema(), required=True)
    comment = fields.Str(required=False, allow_none=True)
    auto_verify = fields.Bool(
        required=False,
        metadata={
            "description": "Verifier choice to auto-verify proof presentation",
            "example": False,
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )


class V10PresentationSendRequestRequestSchema(V10PresentationCreateRequestRequestSchema):
    """Request schema for sending a proof request on a connection."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class V10PresentationSendRequestToProposalSchema(AdminAPIMessageTracingSchema):
    """Request schema for sending a proof request bound to a proposal."""

    auto_verify = fields.Bool(
        required=False,
        metadata={
            "description": "Verifier choice to auto-verify proof presentation",
            "example": False,
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the presentation exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )


class CredentialsFetchQueryStringSchema(OpenAPISchema):
    """Parameters and validators for credentials fetch request query string."""

    referent = fields.Str(
        required=False,
        metadata={
            "description": "Proof request referents of interest, comma-separated",
            "example": "1_name_uuid,2_score_uuid",
        },
    )
    start = fields.Str(
        required=False,
        load_default="0",
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Start index (DEPRECATED - use offset instead)",
            "strict": True,
            "example": NUM_STR_WHOLE_EXAMPLE,
            "deprecated": True,
        },
    )
    count = fields.Str(
        required=False,
        load_default="10",
        validate=NUM_STR_NATURAL_VALIDATE,
        metadata={
            "description": "Maximum number to retrieve (DEPRECATED - use limit instead)",
            "example": NUM_STR_NATURAL_EXAMPLE,
            "deprecated": True,
        },
    )
    limit = fields.Int(
        required=False,
        validate=Range(min=1, max=MAXIMUM_PAGE_SIZE),
        metadata={"description": "Number of results to return", "example": 50},
    )
    offset = fields.Int(
        required=False,
        validate=Range(min=0),
        metadata={"description": "Offset for pagination", "example": 0},
    )
    extra_query = fields.Str(
        required=False,
        validate=INDY_EXTRA_WQL_VALIDATE,
        metadata={
            "description": "(JSON) object mapping referents to extra WQL queries",
            "example": INDY_EXTRA_WQL_EXAMPLE,
        },
    )


class V10PresentationProblemReportRequestSchema(OpenAPISchema):
    """Request schema for sending problem report."""

    description = fields.Str(required=True)


class V10PresExIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking presentation exchange id."""

    pres_ex_id = fields.Str(
        required=True,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Presentation exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )


@docs(
    tags=["present-proof v1.0"],
    summary="Fetch all present-proof exchange records",
    deprecated=True,
)
@querystring_schema(V10PresentationExchangeListQueryStringSchema)
@response_schema(V10PresentationExchangeListSchema(), 200, description="")
@tenant_authentication
async def presentation_exchange_list(request: web.BaseRequest):
    """Request handler for searching presentation exchange records.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange list response

    """
    context: AdminRequestContext = request["context"]
    tag_filter = {}
    if "thread_id" in request.query and request.query["thread_id"] != "":
        tag_filter["thread_id"] = request.query["thread_id"]
    post_filter = {
        k: request.query[k]
        for k in ("connection_id", "role", "state")
        if request.query.get(k, "") != ""
    }

    limit, offset, order_by, descending = get_paginated_query_params(request)

    try:
        async with context.profile.session() as session:
            records = await V10PresentationExchange.query(
                session=session,
                tag_filter=tag_filter,
                limit=limit,
                offset=offset,
                order_by=order_by,
                descending=descending,
                post_filter_positive=post_filter,
            )
        results = [record.serialize() for record in records]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(
    tags=["present-proof v1.0"],
    summary="Fetch a single presentation exchange record",
    deprecated=True,
)
@match_info_schema(V10PresExIdMatchInfoSchema())
@response_schema(V10PresentationExchangeSchema(), 200, description="")
@tenant_authentication
async def presentation_exchange_retrieve(request: web.BaseRequest):
    """Request handler for fetching a single presentation exchange record.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange record response

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    presentation_exchange_id = request.match_info["pres_ex_id"]
    pres_ex_record = None
    try:
        async with profile.session() as session:
            pres_ex_record = await V10PresentationExchange.retrieve_by_id(
                session, presentation_exchange_id
            )
        result = pres_ex_record.serialize()
    except StorageNotFoundError as err:
        # no such pres ex record: not protocol error, user fat-fingered id
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError) as err:
        # present but broken or hopeless: protocol error
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    return web.json_response(result)


@docs(
    tags=["present-proof v1.0"],
    summary="Fetch credentials for a presentation request from wallet",
    deprecated=True,
)
@match_info_schema(V10PresExIdMatchInfoSchema())
@querystring_schema(CredentialsFetchQueryStringSchema())
@response_schema(IndyCredPrecisSchema(many=True), 200, description="")
@tenant_authentication
async def presentation_exchange_credentials_list(request: web.BaseRequest):
    """Request handler for searching applicable credential records.

    Args:
        request: aiohttp request object

    Returns:
        The credential list response

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    presentation_exchange_id = request.match_info["pres_ex_id"]
    referents = request.query.get("referent")
    presentation_referents = (
        (r.strip() for r in referents.split(",")) if referents else ()
    )

    try:
        async with profile.session() as session:
            pres_ex_record = await V10PresentationExchange.retrieve_by_id(
                session, presentation_exchange_id
            )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    # Handle both old style start/count and new limit/offset
    # TODO: Remove start/count and swap to PaginatedQuerySchema and get_limit_offset
    if "limit" in request.query or "offset" in request.query:
        # New style - use limit/offset
        limit = int(request.query.get("limit", DEFAULT_PAGE_SIZE))
        offset = int(request.query.get("offset", 0))
    else:
        # Old style - use start/count
        limit = int(request.query.get("count", "10"))
        offset = int(request.query.get("start", "0"))

    # url encoded json extra_query
    encoded_extra_query = request.query.get("extra_query") or "{}"
    extra_query = json.loads(encoded_extra_query)

    holder = profile.inject(IndyHolder)
    try:
        credentials = await holder.get_credentials_for_presentation_request_by_referent(
            pres_ex_record._presentation_request.ser,
            presentation_referents,
            offset=offset,
            limit=limit,
            extra_query=extra_query,
        )
    except IndyHolderError as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    pres_ex_record.log_state(
        "Retrieved presentation credentials",
        {
            "presentation_exchange_id": presentation_exchange_id,
            "referents": presentation_referents,
            "extra_query": extra_query,
            "credentials": credentials,
        },
        settings=context.settings,
    )
    return web.json_response(credentials)


@docs(
    tags=["present-proof v1.0"],
    summary="Sends a presentation proposal",
    deprecated=True,
)
@request_schema(V10PresentationProposalRequestSchema())
@response_schema(V10PresentationExchangeSchema(), 200, description="")
@tenant_authentication
async def presentation_exchange_send_proposal(request: web.BaseRequest):
    """Request handler for sending a presentation proposal.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    comment = body.get("comment")
    connection_id = body.get("connection_id")

    # Aries RFC 37 calls it a proposal in the proposal struct but it's of type preview
    presentation_preview = body.get("presentation_proposal")
    connection_record = None
    async with profile.session() as session:
        try:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            presentation_proposal_message = PresentationProposal(
                comment=comment,
                presentation_proposal=IndyPresPreview.deserialize(presentation_preview),
            )
        except (BaseModelError, StorageError) as err:
            # other party does not care about our false protocol start
            raise web.HTTPBadRequest(reason=err.roll_up)

    if not connection_record.is_ready:
        raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

    trace_msg = body.get("trace")
    presentation_proposal_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )
    auto_present = body.get(
        "auto_present", context.settings.get("debug.auto_respond_presentation_request")
    )
    auto_remove = body.get("auto_remove")

    presentation_manager = PresentationManager(profile)
    pres_ex_record = None
    try:
        pres_ex_record = await presentation_manager.create_exchange_for_proposal(
            connection_id=connection_id,
            presentation_proposal_message=presentation_proposal_message,
            auto_present=auto_present,
            auto_remove=auto_remove,
        )
        result = pres_ex_record.serialize()
    except (BaseModelError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party does not care about our false protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(presentation_proposal_message, connection_id=connection_id)

    trace_event(
        context.settings,
        presentation_proposal_message,
        outcome="presentation_exchange_propose.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v1.0"],
    summary="Creates a presentation request not bound to any proposal or connection",
    deprecated=True,
)
@request_schema(V10PresentationCreateRequestRequestSchema())
@response_schema(V10PresentationExchangeSchema(), 200, description="")
@tenant_authentication
async def presentation_exchange_create_request(request: web.BaseRequest):
    """Request handler for creating a free presentation request.

    The presentation request will not be bound to any proposal
    or existing connection.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile

    body = await request.json()

    comment = body.get("comment")
    indy_proof_request = body.get("proof_request")
    if not indy_proof_request.get("nonce"):
        indy_proof_request["nonce"] = await generate_pr_nonce()

    presentation_request_message = PresentationRequest(
        comment=comment,
        request_presentations_attach=[
            AttachDecorator.data_base64(
                mapping=indy_proof_request,
                ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
            )
        ],
    )
    auto_verify = body.get(
        "auto_verify", context.settings.get("debug.auto_verify_presentation")
    )
    auto_remove = body.get("auto_remove")
    trace_msg = body.get("trace")
    presentation_request_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )

    pres_ex_record = None
    try:
        presentation_manager = PresentationManager(profile)
        pres_ex_record = await presentation_manager.create_exchange_for_request(
            connection_id=None,
            presentation_request_message=presentation_request_message,
            auto_verify=auto_verify,
            auto_remove=auto_remove,
        )
        result = pres_ex_record.serialize()
    except (BaseModelError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party does not care about our false protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    trace_event(
        context.settings,
        presentation_request_message,
        outcome="presentation_exchange_create_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v1.0"],
    summary="Sends a free presentation request not bound to any proposal",
    deprecated=True,
)
@request_schema(V10PresentationSendRequestRequestSchema())
@response_schema(V10PresentationExchangeSchema(), 200, description="")
@tenant_authentication
async def presentation_exchange_send_free_request(request: web.BaseRequest):
    """Request handler for sending a presentation request free from any proposal.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    async with profile.session() as session:
        try:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
        except StorageNotFoundError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not connection_record.is_ready:
        raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

    comment = body.get("comment")
    indy_proof_request = body.get("proof_request")
    if not indy_proof_request.get("nonce"):
        indy_proof_request["nonce"] = await generate_pr_nonce()

    presentation_request_message = PresentationRequest(
        comment=comment,
        request_presentations_attach=[
            AttachDecorator.data_base64(
                mapping=indy_proof_request,
                ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
            )
        ],
    )
    trace_msg = body.get("trace")
    presentation_request_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )
    auto_verify = body.get(
        "auto_verify", context.settings.get("debug.auto_verify_presentation")
    )
    auto_remove = body.get("auto_remove")

    pres_ex_record = None
    try:
        presentation_manager = PresentationManager(profile)
        pres_ex_record = await presentation_manager.create_exchange_for_request(
            connection_id=connection_id,
            presentation_request_message=presentation_request_message,
            auto_verify=auto_verify,
            auto_remove=auto_remove,
        )
        result = pres_ex_record.serialize()
    except (BaseModelError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party does not care about our false protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(presentation_request_message, connection_id=connection_id)

    trace_event(
        context.settings,
        presentation_request_message,
        outcome="presentation_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v1.0"],
    summary="Sends a presentation request in reference to a proposal",
    deprecated=True,
)
@match_info_schema(V10PresExIdMatchInfoSchema())
@request_schema(V10PresentationSendRequestToProposalSchema())
@response_schema(V10PresentationExchangeSchema(), 200, description="")
@tenant_authentication
async def presentation_exchange_send_bound_request(request: web.BaseRequest):
    """Request handler for sending a presentation request bound to a proposal.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    presentation_exchange_id = request.match_info["pres_ex_id"]
    pres_ex_record = None
    async with profile.session() as session:
        try:
            pres_ex_record = await V10PresentationExchange.retrieve_by_id(
                session, presentation_exchange_id
            )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        if pres_ex_record.state != (V10PresentationExchange.STATE_PROPOSAL_RECEIVED):
            raise web.HTTPBadRequest(
                reason=(
                    f"Presentation exchange {presentation_exchange_id} "
                    f"in {pres_ex_record.state} state "
                    f"(must be {V10PresentationExchange.STATE_PROPOSAL_RECEIVED})"
                )
            )
        conn_id = pres_ex_record.connection_id

        try:
            connection_record = await ConnRecord.retrieve_by_id(session, conn_id)
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not connection_record.is_ready:
        raise web.HTTPForbidden(reason=f"Connection {conn_id} not ready")

    pres_ex_record.auto_verify = body.get(
        "auto_verify", context.settings.get("debug.auto_verify_presentation")
    )
    pres_ex_record.auto_remove = body.get("auto_remove")

    try:
        presentation_manager = PresentationManager(profile)
        (
            pres_ex_record,
            presentation_request_message,
        ) = await presentation_manager.create_bound_request(pres_ex_record)
        result = pres_ex_record.serialize()
    except (BaseModelError, LedgerError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cares that we cannot continue protocol
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    trace_msg = body.get("trace")
    presentation_request_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )
    await outbound_handler(presentation_request_message, connection_id=conn_id)

    trace_event(
        context.settings,
        presentation_request_message,
        outcome="presentation_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v1.0"],
    summary="Sends a proof presentation",
    deprecated=True,
)
@match_info_schema(V10PresExIdMatchInfoSchema())
@request_schema(V10PresentationSendRequestSchema())
@response_schema(V10PresentationExchangeSchema(), description="")
@tenant_authentication
async def presentation_exchange_send_presentation(request: web.BaseRequest):
    """Request handler for sending a presentation.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]
    presentation_exchange_id = request.match_info["pres_ex_id"]
    body = await request.json()

    pres_ex_record = None
    async with profile.session() as session:
        try:
            pres_ex_record = await V10PresentationExchange.retrieve_by_id(
                session, presentation_exchange_id
            )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        if pres_ex_record.state != (V10PresentationExchange.STATE_REQUEST_RECEIVED):
            raise web.HTTPBadRequest(
                reason=(
                    f"Presentation exchange {presentation_exchange_id} "
                    f"in {pres_ex_record.state} state "
                    f"(must be {V10PresentationExchange.STATE_REQUEST_RECEIVED})"
                )
            )

        auto_remove = body.get("auto_remove")
        if auto_remove is None:
            auto_remove = not profile.settings.get("preserve_exchange_records")

        pres_ex_record.auto_remove = auto_remove

        # Fetch connection if exchange has record
        connection_record = None
        if pres_ex_record.connection_id:
            try:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, pres_ex_record.connection_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

    if connection_record and not connection_record.is_ready:
        raise web.HTTPForbidden(
            reason=f"Connection {connection_record.connection_id} not ready"
        )

    try:
        presentation_manager = PresentationManager(profile)
        (
            pres_ex_record,
            presentation_message,
        ) = await presentation_manager.create_presentation(
            pres_ex_record,
            {
                "self_attested_attributes": body.get("self_attested_attributes"),
                "requested_attributes": body.get("requested_attributes"),
                "requested_predicates": body.get("requested_predicates"),
            },
            comment=body.get("comment"),
        )
        result = pres_ex_record.serialize()
    except (
        BaseModelError,
        IndyHolderError,
        LedgerError,
        StorageError,
        WalletNotFoundError,
    ) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cares that we cannot continue protocol
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )

    trace_msg = body.get("trace")
    presentation_message.assign_trace_decorator(
        context.settings,
        trace_msg,
    )
    await outbound_handler(
        presentation_message, connection_id=pres_ex_record.connection_id
    )

    trace_event(
        context.settings,
        presentation_message,
        outcome="presentation_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v1.0"],
    summary="Verify a received presentation",
    deprecated=True,
)
@match_info_schema(V10PresExIdMatchInfoSchema())
@response_schema(V10PresentationExchangeSchema(), description="")
@tenant_authentication
async def presentation_exchange_verify_presentation(request: web.BaseRequest):
    """Request handler for verifying a presentation request.

    Args:
        request: aiohttp request object

    Returns:
        The presentation exchange details

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    presentation_exchange_id = request.match_info["pres_ex_id"]

    pres_ex_record = None
    async with profile.session() as session:
        try:
            pres_ex_record = await V10PresentationExchange.retrieve_by_id(
                session, presentation_exchange_id
            )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        if pres_ex_record.state != (V10PresentationExchange.STATE_PRESENTATION_RECEIVED):
            raise web.HTTPBadRequest(
                reason=(
                    f"Presentation exchange {presentation_exchange_id} "
                    f"in {pres_ex_record.state} state "
                    f"(must be {V10PresentationExchange.STATE_PRESENTATION_RECEIVED})"
                )
            )

    try:
        presentation_manager = PresentationManager(profile)
        pres_ex_record = await presentation_manager.verify_presentation(pres_ex_record)
        result = pres_ex_record.serialize()
    except (BaseModelError, LedgerError, StorageError) as err:
        if pres_ex_record:
            async with profile.session() as session:
                await pres_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cares that we cannot continue protocol
        await report_problem(
            err,
            ProblemReportReason.ABANDONED.value,
            web.HTTPBadRequest,
            pres_ex_record,
            outbound_handler,
        )
    except PresentationManagerError as err:
        return web.HTTPBadRequest(reason=err.roll_up)

    trace_event(
        context.settings,
        pres_ex_record,
        outcome="presentation_exchange_verify.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["present-proof v1.0"],
    summary="Send a problem report for presentation exchange",
    deprecated=True,
)
@match_info_schema(V10PresExIdMatchInfoSchema())
@request_schema(V10PresentationProblemReportRequestSchema())
@response_schema(V10PresentProofModuleResponseSchema(), 200, description="")
@tenant_authentication
async def presentation_exchange_problem_report(request: web.BaseRequest):
    """Request handler for sending problem report.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    pres_ex_id = request.match_info["pres_ex_id"]
    body = await request.json()
    description = body["description"]

    try:
        async with await context.profile.session() as session:
            pres_ex_record = await V10PresentationExchange.retrieve_by_id(
                session, pres_ex_id
            )
            report = problem_report_for_record(pres_ex_record, description)
            await pres_ex_record.save_error_state(
                session,
                reason=f"created problem report: {description}",
            )
    except StorageNotFoundError as err:  # other party does not care about meta-problems
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(report, connection_id=pres_ex_record.connection_id)

    return web.json_response({})


@docs(
    tags=["present-proof v1.0"],
    summary="Remove an existing presentation exchange record",
    deprecated=True,
)
@match_info_schema(V10PresExIdMatchInfoSchema())
@response_schema(V10PresentProofModuleResponseSchema(), description="")
@tenant_authentication
async def presentation_exchange_remove(request: web.BaseRequest):
    """Request handler for removing a presentation exchange record.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]

    presentation_exchange_id = request.match_info["pres_ex_id"]
    pres_ex_record = None
    try:
        async with context.profile.session() as session:
            pres_ex_record = await V10PresentationExchange.retrieve_by_id(
                session, presentation_exchange_id
            )
            await pres_ex_record.delete_record(session)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get(
                "/present-proof/records",
                presentation_exchange_list,
                allow_head=False,
            ),
            web.get(
                "/present-proof/records/{pres_ex_id}",
                presentation_exchange_retrieve,
                allow_head=False,
            ),
            web.get(
                "/present-proof/records/{pres_ex_id}/credentials",
                presentation_exchange_credentials_list,
                allow_head=False,
            ),
            web.post(
                "/present-proof/send-proposal",
                presentation_exchange_send_proposal,
            ),
            web.post(
                "/present-proof/create-request",
                presentation_exchange_create_request,
            ),
            web.post(
                "/present-proof/send-request",
                presentation_exchange_send_free_request,
            ),
            web.post(
                "/present-proof/records/{pres_ex_id}/send-request",
                presentation_exchange_send_bound_request,
            ),
            web.post(
                "/present-proof/records/{pres_ex_id}/send-presentation",
                presentation_exchange_send_presentation,
            ),
            web.post(
                "/present-proof/records/{pres_ex_id}/verify-presentation",
                presentation_exchange_verify_presentation,
            ),
            web.post(
                "/present-proof/records/{pres_ex_id}/problem-report",
                presentation_exchange_problem_report,
            ),
            web.delete(
                "/present-proof/records/{pres_ex_id}",
                presentation_exchange_remove,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "present-proof v1.0",
            "description": "Proof presentation v1.0",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""DID rotate admin routes."""

import logging

from aiohttp import web
from aiohttp_apispec import docs, json_schema, match_info_schema, response_schema
from marshmallow import fields

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import DID_WEB_EXAMPLE, UUID4_EXAMPLE
from ....storage.error import StorageNotFoundError
from .manager import (
    DIDRotateManager,
    UnresolvableDIDCommServicesError,
    UnresolvableDIDError,
    UnsupportedDIDMethodError,
)
from .message_types import SPEC_URI
from .messages.hangup import HangupSchema as HangupMessageSchema
from .messages.rotate import RotateSchema as RotateMessageSchema

LOGGER = logging.getLogger(__name__)


class DIDRotateConnIdMatchInfoSchema(OpenAPISchema):
    """Request to rotate a DID."""

    conn_id = fields.String(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class DIDRotateRequestJSONSchema(OpenAPISchema):
    """Request to rotate a DID."""

    to_did = fields.String(
        required=True,
        metadata={
            "description": "The DID the rotating party is rotating to",
            "example": DID_WEB_EXAMPLE,
        },
    )


@docs(tags=["did-rotate"], summary="Begin rotation of a DID as a rotator")
@match_info_schema(DIDRotateConnIdMatchInfoSchema())
@json_schema(DIDRotateRequestJSONSchema())
@response_schema(
    RotateMessageSchema(), 200, description="Rotate agent message for observer"
)
@tenant_authentication
async def rotate(request: web.BaseRequest):
    """Request to rotate a DID."""

    LOGGER.debug("DID Rotate Rotate request >>>")

    context: AdminRequestContext = request["context"]

    profile = context.profile
    did_rotate_mgr = DIDRotateManager(profile)

    connection_id = request.match_info["conn_id"]

    body = await request.json()
    to_did = body["to_did"]

    # Validate DID before proceeding
    try:
        await did_rotate_mgr.ensure_supported_did(to_did)
    except (
        UnsupportedDIDMethodError,
        UnresolvableDIDError,
        UnresolvableDIDCommServicesError,
    ) as err:
        raise web.HTTPBadRequest(reason=str(err)) from err

    async with context.session() as session:
        try:
            conn = await ConnRecord.retrieve_by_id(session, connection_id)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        msg = await did_rotate_mgr.rotate_my_did(conn, to_did)

        return web.json_response(msg.serialize())


@docs(tags=["did-rotate"], summary="Send hangup of DID rotation as a rotator")
@match_info_schema(DIDRotateConnIdMatchInfoSchema())
@response_schema(
    HangupMessageSchema(), 200, description="Hangup agent message for observer"
)
@tenant_authentication
async def hangup(request: web.BaseRequest):
    """Hangup a DID rotation."""

    LOGGER.debug("DID Rotate Hangup request >>>")

    context: AdminRequestContext = request["context"]

    profile = context.profile
    did_rotate_mgr = DIDRotateManager(profile)

    connection_id = request.match_info["conn_id"]

    async with context.session() as session:
        try:
            conn = await ConnRecord.retrieve_by_id(session, connection_id)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        msg = await did_rotate_mgr.hangup(conn)

        return web.json_response(msg.serialize())


async def register(app: web.Application):
    """Register routes."""

    app.add_routes([web.post("/did-rotate/{conn_id}/rotate", rotate)])
    app.add_routes([web.post("/did-rotate/{conn_id}/hangup", hangup)])


def post_process_routes(app: web.Application):
    """Amend Swagger API."""

    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "did-rotate",
            "description": "Rotate a DID",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""Transaction resend handler."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..manager import TransactionManager, TransactionManagerError
from ..messages.transaction_resend import TransactionResend


class TransactionResendHandler(BaseHandler):
    """Handler class for transaction resend."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle transaction resend.

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug("TransactionResendHandler called with context %s", context)
        assert isinstance(context.message, TransactionResend)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        mgr = TransactionManager(context.profile)
        try:
            await mgr.receive_transaction_resend(
                context.message, context.connection_record.connection_id
            )
        except TransactionManagerError:
            self._logger.exception("Error receiving resend transaction request")



"""MultiBase encoding and decoding utilities."""

from abc import ABC, abstractmethod
from enum import Enum
from typing import ClassVar, Literal, Union


class MultibaseEncoder(ABC):
    """Encoding details."""

    name: ClassVar[str]
    character: ClassVar[str]

    @abstractmethod
    def encode(self, value: bytes) -> str:
        """Encode a byte string using this encoding."""

    @abstractmethod
    def decode(self, value: str) -> bytes:
        """Decode a string using this encoding."""


class Base58BtcEncoder(MultibaseEncoder):
    """Base58BTC encoding."""

    name = "base58btc"
    character = "z"

    def encode(self, value: bytes) -> str:
        """Encode a byte string using the base58btc encoding."""
        import base58

        return base58.b58encode(value).decode()

    def decode(self, value: str) -> bytes:
        """Decode a multibase encoded string."""
        import base58

        return base58.b58decode(value)


class Encoding(Enum):
    """Enum for supported encodings."""

    base58btc = Base58BtcEncoder()
    # Insert additional encodings here

    @classmethod
    def from_name(cls, name: str) -> MultibaseEncoder:
        """Get encoding from name."""
        for encoding in cls:
            if encoding.value.name == name:
                return encoding.value
        raise ValueError(f"Unsupported encoding: {name}")

    @classmethod
    def from_character(cls, character: str) -> MultibaseEncoder:
        """Get encoding from character."""
        for encoding in cls:
            if encoding.value.character == character:
                return encoding.value
        raise ValueError(f"Unsupported encoding: {character}")


EncodingStr = Literal[
    "base58btc",
    # Insert additional encoding names here
]


def encode(value: bytes, encoding: Union[Encoding, EncodingStr]) -> str:
    """Encode a byte string using the given encoding.

    Args:
        value: The byte string to encode
        encoding: The encoding to use

    Returns:
        The encoded string
    """
    if isinstance(encoding, str):
        encoder = Encoding.from_name(encoding)
    elif isinstance(encoding, Encoding):
        encoder = encoding.value
    else:
        raise TypeError("encoding must be an Encoding or EncodingStr")

    return encoder.character + encoder.encode(value)


def decode(value: str) -> bytes:
    """Decode a multibase encoded string.

    Args:
        value: The string to decode

    Returns:
        The decoded byte string
    """
    encoding = value[0]
    encoded = value[1:]
    encoder = Encoding.from_character(encoding)

    return encoder.decode(encoded)



import pytest
import pytest_asyncio

from ......core.protocol_registry import ProtocolRegistry
from ......messaging.base_handler import HandlerException
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......storage.error import StorageNotFoundError
from ......tests import mock
from ......utils.testing import create_test_profile
from .....didcomm_prefix import DIDCommPrefix
from ...handlers.disclosures_handler import DisclosuresHandler
from ...messages.disclosures import Disclosures
from ...messages.queries import Queries, QueryItem
from ...models.discovery_record import V20DiscoveryExchangeRecord

TEST_MESSAGE_FAMILY = "doc/proto/1.0"
TEST_MESSAGE_TYPE = TEST_MESSAGE_FAMILY + "/message"


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    ctx.connection_ready = True
    ctx.connection_record = mock.MagicMock(connection_id="test123")
    yield ctx


class TestDisclosuresHandler:
    @pytest.mark.asyncio
    async def test_disclosures(self, request_context):
        registry = ProtocolRegistry()
        registry.register_message_types({TEST_MESSAGE_TYPE: object()})
        request_context.injector.bind_instance(ProtocolRegistry, registry)
        disclosures = Disclosures(
            disclosures=[
                {
                    "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
                    "feature-type": "protocol",
                    "roles": [],
                },
                {"feature-type": "goal-code", "id": "aries.sell.goods.consumer"},
            ]
        )
        test_queries = [
            QueryItem(feature_type="protocol", match="https://didcomm.org/tictactoe/1.*"),
            QueryItem(feature_type="goal-code", match="aries.*"),
        ]
        queries = Queries(queries=test_queries)
        discovery_record = V20DiscoveryExchangeRecord(
            connection_id="test123",
            thread_id="test123",
            queries_msg=queries,
        )
        disclosures.assign_thread_id("test123")
        request_context.message = disclosures

        handler = DisclosuresHandler()
        mock_responder = MockResponder()
        with mock.patch.object(
            V20DiscoveryExchangeRecord,
            "retrieve_by_id",
            mock.CoroutineMock(return_value=discovery_record),
        ):
            await handler.handle(request_context, mock_responder)
            assert not mock_responder.messages

    @pytest.mark.asyncio
    async def test_disclosures_connection_id_no_thid(self, request_context):
        registry = ProtocolRegistry()
        registry.register_message_types({TEST_MESSAGE_TYPE: object()})
        request_context.injector.bind_instance(ProtocolRegistry, registry)
        disclosures = Disclosures(
            disclosures=[
                {
                    "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
                    "feature-type": "protocol",
                    "roles": [],
                },
                {"feature-type": "goal-code", "id": "aries.sell.goods.consumer"},
            ]
        )
        test_queries = [
            QueryItem(feature_type="protocol", match="https://didcomm.org/tictactoe/1.*"),
            QueryItem(feature_type="goal-code", match="aries.*"),
        ]
        queries = Queries(queries=test_queries)
        discovery_record = V20DiscoveryExchangeRecord(
            connection_id="test123",
            thread_id="test123",
            queries_msg=queries,
        )
        disclosures.assign_thread_id("test123")
        request_context.message = disclosures

        handler = DisclosuresHandler()
        mock_responder = MockResponder()
        with (
            mock.patch.object(
                V20DiscoveryExchangeRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=StorageNotFoundError),
            ),
            mock.patch.object(
                V20DiscoveryExchangeRecord,
                "retrieve_by_connection_id",
                mock.CoroutineMock(return_value=discovery_record),
            ),
        ):
            await handler.handle(request_context, mock_responder)
            assert not mock_responder.messages

    @pytest.mark.asyncio
    async def test_disclosures_no_conn_id_no_thid(self, request_context):
        registry = ProtocolRegistry()
        registry.register_message_types({TEST_MESSAGE_TYPE: object()})
        request_context.injector.bind_instance(ProtocolRegistry, registry)
        disclosures = Disclosures(
            disclosures=[
                {
                    "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
                    "feature-type": "protocol",
                    "roles": [],
                },
                {"feature-type": "goal-code", "id": "aries.sell.goods.consumer"},
            ]
        )

        disclosures.assign_thread_id("test123")
        request_context.message = disclosures

        handler = DisclosuresHandler()
        mock_responder = MockResponder()
        with (
            mock.patch.object(
                V20DiscoveryExchangeRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=StorageNotFoundError),
            ),
            mock.patch.object(
                V20DiscoveryExchangeRecord,
                "retrieve_by_connection_id",
                mock.CoroutineMock(side_effect=StorageNotFoundError),
            ),
        ):
            await handler.handle(request_context, mock_responder)
            assert not mock_responder.messages

    @pytest.mark.asyncio
    async def test_disclose_connection_not_ready(self, request_context):
        request_context.connection_ready = False
        disclosures = Disclosures(
            disclosures=[
                {
                    "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
                    "feature-type": "protocol",
                    "roles": [],
                },
                {"feature-type": "goal-code", "id": "aries.sell.goods.consumer"},
            ]
        )
        disclosures.assign_thread_id("test123")
        request_context.message = disclosures
        handler = DisclosuresHandler()
        mock_responder = MockResponder()
        with pytest.raises(HandlerException):
            await handler.handle(request_context, mock_responder)



"""Test Base DID Resolver methods."""

import re
from unittest import mock

import pytest
from pydid import DIDDocument

from ..base import BaseDIDResolver, DIDMethodNotSupported, ResolverType


class ExampleDIDResolver(BaseDIDResolver):
    """Test DID Resolver."""

    def __init__(self):
        super().__init__()

    async def setup(self, context):
        pass

    @property
    def supported_did_regex(self):
        return re.compile("^did:example:[a-zA-Z0-9_.-]+$")

    async def _resolve(self, profile, did, accept) -> DIDDocument:
        return DIDDocument("did:example:123")


@pytest.fixture
def native_resolver():
    resolver = ExampleDIDResolver()
    resolver.type = ResolverType.NATIVE
    yield resolver


@pytest.fixture
def non_native_resolver():
    yield ExampleDIDResolver()


@pytest.fixture
def profile():
    yield mock.MagicMock()


def test_native_on_native(native_resolver):
    assert native_resolver.native is True


def test_native_on_non_native(non_native_resolver):
    assert non_native_resolver.native is False


@pytest.mark.asyncio
async def test_supports(profile, native_resolver):
    assert not await native_resolver.supports(profile, "did:test:basdfasdfas")
    assert await native_resolver.supports(profile, "did:example:WgWxqztrNooG92RXvxSTWv")


@pytest.mark.asyncio
async def test_resolve_x(native_resolver):
    with pytest.raises(DIDMethodNotSupported) as x_did:
        await native_resolver.resolve(None, "did:nosuchmethod:xxx")
    assert "does not support DID method" in str(x_did.value)


@pytest.mark.asyncio
async def test_supported_methods():
    class TestDIDResolver(BaseDIDResolver):
        async def setup(self, context):
            pass

        @property
        def supported_methods(self):
            return ["example"]

        async def _resolve(self, profile, did, accept) -> DIDDocument:
            return DIDDocument("did:example:123")

    with pytest.deprecated_call():
        assert await TestDIDResolver().supports(
            profile, "did:example:WgWxqztrNooG92RXvxSTWv"
        )



"""AnonCreds admin routes."""

import logging
from asyncio import shield

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..core.event_bus import EventBus
from ..messaging.models.openapi import OpenAPISchema
from ..messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_DID_EXAMPLE,
    ANONCREDS_REV_REG_ID_EXAMPLE,
    ANONCREDS_REV_REG_ID_VALIDATE,
    ANONCREDS_SCHEMA_ID_EXAMPLE,
    UUIDFour,
)
from ..revocation.error import RevocationNotSupportedError
from ..storage.error import StorageNotFoundError
from ..utils.profiles import is_not_anoncreds_profile_raise_web_exception
from .base import (
    AnonCredsObjectNotFound,
    AnonCredsRegistrationError,
    AnonCredsResolutionError,
)
from .issuer import AnonCredsIssuer, AnonCredsIssuerError
from .models.credential_definition import CredDefResultSchema, GetCredDefResultSchema
from .models.revocation import RevListResultSchema, RevRegDefResultSchema
from .models.schema import (
    AnonCredsSchemaSchema,
    GetSchemaResultSchema,
    SchemaResultSchema,
)
from .registry import AnonCredsRegistry
from .revocation import AnonCredsRevocation, AnonCredsRevocationError
from .revocation_setup import DefaultRevocationSetup
from .util import handle_value_error

LOGGER = logging.getLogger(__name__)

CRED_DEF_TAG_TITLE = "AnonCreds - Credential Definitions"
SCHEMAS_TAG_TITLE = "AnonCreds - Schemas"
REVOCATION_TAG_TITLE = "AnonCreds - Revocation"

SPEC_URI = "https://hyperledger.github.io/anoncreds-spec"

endorser_connection_id_description = (
    "Connection identifier (optional) (this is an example). "
    "You can set this if you know the endorser's connection id you want to use. "
    "If not specified then the agent will attempt to find an endorser connection."
)
create_transaction_for_endorser_description = (
    "Create transaction for endorser (optional, default false). "
    "Use this for agents who don't specify an author role but want to "
    "create a transaction for an endorser to sign."
)


class AnonCredsRevocationModuleResponseSchema(OpenAPISchema):
    """Response schema for Revocation Module."""


class AnonCredsRevRegIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking rev reg id."""

    rev_reg_id = fields.Str(
        required=True,
        validate=ANONCREDS_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation Registry identifier",
            "example": ANONCREDS_REV_REG_ID_EXAMPLE,
        },
    )


class SchemaIdMatchInfo(OpenAPISchema):
    """Path parameters and validators for request taking schema id."""

    schema_id = fields.Str(
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        }
    )


class SchemaPostOptionSchema(OpenAPISchema):
    """Parameters and validators for schema options."""

    endorser_connection_id = fields.Str(
        metadata={
            "description": endorser_connection_id_description,
            "example": UUIDFour.EXAMPLE,
        },
        required=False,
    )

    create_transaction_for_endorser = fields.Bool(
        metadata={
            "description": create_transaction_for_endorser_description,
            "example": False,
        },
        required=False,
    )


class SchemasQueryStringSchema(OpenAPISchema):
    """Parameters and validators for query string in schemas list query."""

    schema_name = fields.Str(
        metadata={
            "description": "Schema name",
            "example": "example-schema",
        }
    )
    schema_version = fields.Str(
        metadata={
            "description": "Schema version",
            "example": "1.0",
        }
    )
    schema_issuer_id = fields.Str(
        metadata={
            "description": "Schema issuer identifier",
            "example": ANONCREDS_DID_EXAMPLE,
        }
    )


class GetSchemasResponseSchema(OpenAPISchema):
    """Parameters and validators for schema list all response."""

    schema_ids = fields.List(
        fields.Str(
            metadata={
                "description": "Schema identifiers",
                "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
            }
        )
    )


class SchemaPostRequestSchema(OpenAPISchema):
    """Parameters and validators for query string in create schema."""

    schema = fields.Nested(AnonCredsSchemaSchema())
    options = fields.Nested(SchemaPostOptionSchema())


@docs(
    tags=[SCHEMAS_TAG_TITLE],
    summary="Create a schema on the connected datastore",
)
@request_schema(SchemaPostRequestSchema())
@response_schema(SchemaResultSchema(), 200, description="")
@tenant_authentication
async def schemas_post(request: web.BaseRequest):
    """Request handler for creating a schema.

    Args:
        request (web.BaseRequest): aiohttp request object
          schema: {
                "attrNames": ["string"],
                "name": "string",
                "version": "string",
                "issuerId": "string"
            },
        options: options method can be different per method,
            but it can also include default options for all anoncreds
            methods (none for schema). it can also be automatically
            inferred from the agent startup parameters (default endorser)
            endorser_connection_id: ""
    Returns:
        json object:
            job_id: job identifier to keep track of the status of the schema creation.
            MUST be absent or have a null value if the value of the schema_state. state
            response field is either finished or failed, and MUST NOT have a null value
            otherwise.
            schema_state:
                state : The state of the schema creation. Possible values are finished,
                failed, action and wait.
                schema_id : The id of the schema. If the value of the schema_state.state
                response field is finished, this field MUST be present and MUST NOT have
                a null value.
                schema : The schema. If the value of the schema_state.state response field
                is finished, this field MUST be present and MUST NOT have a null value.
            registration_metadata : This field contains metadata about the registration
            process
            schema_metadata : This fields contains metadata about the schema.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    options = body.get("options", {})
    schema_data = body.get("schema")

    if schema_data is None:
        raise web.HTTPBadRequest(reason="schema object is required")

    issuer_id = schema_data.get("issuerId")
    attr_names = schema_data.get("attrNames")
    name = schema_data.get("name")
    version = schema_data.get("version")

    try:
        issuer = AnonCredsIssuer(profile)
        result = await issuer.create_and_register_schema(
            issuer_id,
            name,
            version,
            attr_names,
            options,
        )
        return web.json_response(result.serialize())
    except ValueError as e:
        handle_value_error(e)
    except (AnonCredsIssuerError, AnonCredsRegistrationError) as e:
        raise web.HTTPBadRequest(reason=e.roll_up) from e


@docs(
    tags=[SCHEMAS_TAG_TITLE],
    summary="Retrieve an individual schemas details",
)
@match_info_schema(SchemaIdMatchInfo())
@response_schema(GetSchemaResultSchema(), 200, description="")
@tenant_authentication
async def schema_get(request: web.BaseRequest):
    """Request handler for getting a schema.

    Args:
        request (web.BaseRequest): aiohttp request object

    Returns:
        json object: schema

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    anoncreds_registry = context.inject(AnonCredsRegistry)
    schema_id = request.match_info["schema_id"]
    try:
        schema = await anoncreds_registry.get_schema(profile, schema_id)
        return web.json_response(schema.serialize())
    except AnonCredsObjectNotFound as e:
        raise web.HTTPNotFound(reason=f"Schema not found: {schema_id}") from e
    except AnonCredsResolutionError as e:
        raise web.HTTPBadRequest(reason=e.roll_up) from e


@docs(
    tags=[SCHEMAS_TAG_TITLE],
    summary="Retrieve all schema ids",
)
@querystring_schema(SchemasQueryStringSchema())
@response_schema(GetSchemasResponseSchema(), 200, description="")
@tenant_authentication
async def schemas_get(request: web.BaseRequest):
    """Request handler for getting all schemas.

    Args:
        request: aiohttp request object

    Returns:
        The credential definition details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    schema_issuer_id = request.query.get("schema_issuer_id")
    schema_name = request.query.get("schema_name")
    schema_version = request.query.get("schema_version")

    try:
        issuer = AnonCredsIssuer(profile)
        schema_ids = await issuer.get_created_schemas(
            schema_name, schema_version, schema_issuer_id
        )
    except ValueError as e:
        handle_value_error(e)
    return web.json_response({"schema_ids": schema_ids})


class CredIdMatchInfo(OpenAPISchema):
    """Path parameters and validators for request taking credential id."""

    cred_def_id = fields.Str(
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
        required=True,
    )


class InnerCredDefSchema(OpenAPISchema):
    """Parameters and validators for credential definition."""

    tag = fields.Str(
        metadata={
            "description": "Credential definition tag",
            "example": "default",
        },
        required=True,
    )
    schema_id = fields.Str(
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        },
        required=True,
        data_key="schemaId",
    )
    issuer_id = fields.Str(
        metadata={
            "description": "Issuer Identifier of the credential definition",
            "example": ANONCREDS_DID_EXAMPLE,
        },
        required=True,
        data_key="issuerId",
    )


class CredDefPostOptionsSchema(OpenAPISchema):
    """Parameters and validators for credential definition options."""

    endorser_connection_id = fields.Str(
        metadata={
            "description": endorser_connection_id_description,
            "example": UUIDFour.EXAMPLE,
        },
        required=False,
    )
    create_transaction_for_endorser = fields.Bool(
        metadata={
            "description": create_transaction_for_endorser_description,
            "example": False,
        },
        required=False,
    )
    support_revocation = fields.Bool(
        metadata={
            "description": "Support credential revocation",
        },
        required=False,
    )
    revocation_registry_size = fields.Int(
        metadata={
            "description": "Maximum number of credential revocations per registry",
            "example": 1000,
        },
        required=False,
    )


class CredDefPostRequestSchema(OpenAPISchema):
    """Parameters and validators for query string in create credential definition."""

    credential_definition = fields.Nested(InnerCredDefSchema())
    options = fields.Nested(CredDefPostOptionsSchema())


class CredDefsQueryStringSchema(OpenAPISchema):
    """Parameters and validators for credential definition list query."""

    issuer_id = fields.Str(
        metadata={
            "description": "Issuer Identifier of the credential definition",
            "example": ANONCREDS_DID_EXAMPLE,
        }
    )
    schema_id = fields.Str(
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        }
    )
    schema_name = fields.Str(
        metadata={
            "description": "Schema name",
            "example": "example-schema",
        }
    )
    schema_version = fields.Str(
        metadata={
            "description": "Schema version",
            "example": "1.0",
        }
    )


@docs(
    tags=[CRED_DEF_TAG_TITLE],
    summary="Create a credential definition on the connected datastore",
)
@request_schema(CredDefPostRequestSchema())
@response_schema(CredDefResultSchema(), 200, description="")
@tenant_authentication
async def cred_def_post(request: web.BaseRequest):
    """Request handler for creating .

    Args:
        request: aiohttp request object

    Returns:
        The credential definition details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    options = body.get("options", {})
    cred_def = body.get("credential_definition")

    if cred_def is None:
        raise web.HTTPBadRequest(reason="cred_def object is required")

    issuer_id = cred_def.get("issuerId")
    schema_id = cred_def.get("schemaId")
    tag = cred_def.get("tag")

    try:
        issuer = AnonCredsIssuer(profile)
        result = await issuer.create_and_register_credential_definition(
            issuer_id,
            schema_id,
            tag,
            options=options,
        )
        return web.json_response(result.serialize())
    except ValueError as e:
        handle_value_error(e)
    except (
        AnonCredsIssuerError,
        AnonCredsObjectNotFound,
        AnonCredsResolutionError,
    ) as e:
        raise web.HTTPBadRequest(reason=e.roll_up) from e


@docs(
    tags=[CRED_DEF_TAG_TITLE],
    summary="Retrieve an individual credential definition details",
)
@match_info_schema(CredIdMatchInfo())
@response_schema(GetCredDefResultSchema(), 200, description="")
@tenant_authentication
async def cred_def_get(request: web.BaseRequest):
    """Request handler for getting credential definition.

    Args:
        request: aiohttp request object

    Returns:
        The credential definition details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    anon_creds_registry = context.inject(AnonCredsRegistry)
    credential_id = request.match_info["cred_def_id"]
    try:
        result = await anon_creds_registry.get_credential_definition(
            profile, credential_id
        )
        return web.json_response(result.serialize())
    except AnonCredsObjectNotFound as e:
        raise web.HTTPBadRequest(
            reason=f"Credential definition {credential_id} not found"
        ) from e


class GetCredDefsResponseSchema(OpenAPISchema):
    """AnonCredsRegistryGetCredDefsSchema."""

    credential_definition_ids = fields.List(
        fields.Str(
            metadata={
                "description": "credential definition identifiers",
                "example": "GvLGiRogTJubmj5B36qhYz:3:CL:8:faber.agent.degree_schema",
            }
        )
    )


@docs(
    tags=[CRED_DEF_TAG_TITLE],
    summary="Retrieve all credential definition ids",
)
@querystring_schema(CredDefsQueryStringSchema())
@response_schema(GetCredDefsResponseSchema(), 200, description="")
@tenant_authentication
async def cred_defs_get(request: web.BaseRequest):
    """Request handler for getting all credential definitions.

    Args:
        request: aiohttp request object

    Returns:
        The credential definition details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    try:
        issuer = AnonCredsIssuer(profile)

        cred_def_ids = await issuer.get_created_credential_definitions(
            issuer_id=request.query.get("issuer_id"),
            schema_id=request.query.get("schema_id"),
            schema_name=request.query.get("schema_name"),
            schema_version=request.query.get("schema_version"),
        )
        return web.json_response({"credential_definition_ids": cred_def_ids})
    except ValueError as e:
        handle_value_error(e)


class InnerRevRegDefSchema(OpenAPISchema):
    """Request schema for revocation registry creation request."""

    issuer_id = fields.Str(
        metadata={
            "description": "Issuer Identifier of the credential definition or schema",
            "example": ANONCREDS_DID_EXAMPLE,
        },
        data_key="issuerId",
        required=True,
    )
    cred_def_id = fields.Str(
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        },
        data_key="credDefId",
        required=True,
    )
    tag = fields.Str(
        metadata={"description": "tag for revocation registry", "example": "default"},
        required=True,
    )
    max_cred_num = fields.Int(
        metadata={
            "description": "Maximum number of credential revocations per registry",
            "example": 777,
        },
        data_key="maxCredNum",
        required=True,
    )


class RevRegDefOptionsSchema(OpenAPISchema):
    """Parameters and validators for rev reg def options."""

    endorser_connection_id = fields.Str(
        metadata={
            "description": endorser_connection_id_description,
            "example": UUIDFour.EXAMPLE,
        },
        required=False,
    )
    create_transaction_for_endorser = fields.Bool(
        metadata={
            "description": create_transaction_for_endorser_description,
            "example": False,
        },
        required=False,
    )


class RevRegCreateRequestSchemaAnonCreds(OpenAPISchema):
    """Wrapper for revocation registry creation request."""

    revocation_registry_definition = fields.Nested(InnerRevRegDefSchema())
    options = fields.Nested(RevRegDefOptionsSchema())


@docs(
    tags=[REVOCATION_TAG_TITLE],
    summary="Create and publish a registration revocation on the connected datastore",
)
@request_schema(RevRegCreateRequestSchemaAnonCreds())
@response_schema(RevRegDefResultSchema(), 200, description="")
@tenant_authentication
async def rev_reg_def_post(request: web.BaseRequest):
    """Request handler for creating revocation registry definition."""
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    revocation_registry_definition = body.get("revocation_registry_definition")
    options = body.get("options", {})

    if revocation_registry_definition is None:
        raise web.HTTPBadRequest(
            reason="revocation_registry_definition object is required"
        )

    issuer_id = revocation_registry_definition.get("issuerId")
    cred_def_id = revocation_registry_definition.get("credDefId")
    max_cred_num = revocation_registry_definition.get("maxCredNum")
    tag = revocation_registry_definition.get("tag")

    issuer = AnonCredsIssuer(profile)
    revocation = AnonCredsRevocation(profile)
    # check we published this cred def
    found = await issuer.match_created_credential_definitions(cred_def_id)
    if not found:
        raise web.HTTPNotFound(
            reason=f"Not issuer of credential definition id {cred_def_id}"
        )

    try:
        result = await shield(
            revocation.create_and_register_revocation_registry_definition(
                issuer_id,
                cred_def_id,
                registry_type="CL_ACCUM",
                max_cred_num=max_cred_num,
                tag=tag,
                options=options,
            )
        )
        return web.json_response(result.serialize())
    except (RevocationNotSupportedError, AnonCredsRevocationError) as e:
        raise web.HTTPBadRequest(reason=e.roll_up) from e


class RevListOptionsSchema(OpenAPISchema):
    """Parameters and validators for revocation list options."""

    endorser_connection_id = fields.Str(
        metadata={
            "description": endorser_connection_id_description,
            "example": UUIDFour.EXAMPLE,
        },
        required=False,
    )
    create_transaction_for_endorser = fields.Bool(
        metadata={
            "description": create_transaction_for_endorser_description,
            "example": False,
        },
        required=False,
    )


class RevListCreateRequestSchema(OpenAPISchema):
    """Request schema for revocation registry creation request."""

    rev_reg_def_id = fields.Str(
        metadata={
            "description": "Revocation registry definition identifier",
            "example": ANONCREDS_REV_REG_ID_EXAMPLE,
        },
        required=True,
    )
    options = fields.Nested(RevListOptionsSchema)


@docs(
    tags=[REVOCATION_TAG_TITLE],
    summary="Create and publish a revocation status list on the connected datastore",
)
@request_schema(RevListCreateRequestSchema())
@response_schema(RevListResultSchema(), 200, description="")
@tenant_authentication
async def rev_list_post(request: web.BaseRequest):
    """Request handler for creating registering a revocation list."""
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    rev_reg_def_id = body.get("rev_reg_def_id")
    options = body.get("options", {})

    try:
        revocation = AnonCredsRevocation(profile)
        result = await shield(
            revocation.create_and_register_revocation_list(
                rev_reg_def_id,
                options,
            )
        )
        LOGGER.debug("published revocation list for: %s", rev_reg_def_id)
        return web.json_response(result.serialize())
    except ValueError as e:
        handle_value_error(e)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except AnonCredsRevocationError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err


@docs(
    tags=[REVOCATION_TAG_TITLE],
    summary="Upload local tails file to server",
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@response_schema(AnonCredsRevocationModuleResponseSchema(), description="")
@tenant_authentication
async def upload_tails_file(request: web.BaseRequest):
    """Request handler to upload local tails file for revocation registry.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    try:
        revocation = AnonCredsRevocation(profile)
        rev_reg_def = await revocation.get_created_revocation_registry_definition(
            rev_reg_id
        )
        if rev_reg_def is None:
            raise web.HTTPNotFound(reason="No rev reg def found")

        await revocation.upload_tails_file(rev_reg_def)
        return web.json_response({})
    except ValueError as e:
        handle_value_error(e)
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e


@docs(
    tags=[REVOCATION_TAG_TITLE],
    summary="Update the active registry",
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@response_schema(AnonCredsRevocationModuleResponseSchema(), description="")
@tenant_authentication
async def set_active_registry(request: web.BaseRequest):
    """Request handler to set the active registry.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    try:
        revocation = AnonCredsRevocation(profile)
        await revocation.set_active_registry(rev_reg_id)
        return web.json_response({})
    except ValueError as e:
        handle_value_error(e)
    except AnonCredsRevocationError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e


def register_events(event_bus: EventBus) -> None:
    """Register events."""
    # TODO Make this pluggable?
    setup_manager = DefaultRevocationSetup()
    setup_manager.register_events(event_bus)


async def register(app: web.Application) -> None:
    """Register routes."""

    app.add_routes(
        [
            web.post("/anoncreds/schema", schemas_post),
            web.get("/anoncreds/schema/{schema_id}", schema_get, allow_head=False),
            web.get("/anoncreds/schemas", schemas_get, allow_head=False),
            web.post("/anoncreds/credential-definition", cred_def_post),
            web.get(
                "/anoncreds/credential-definition/{cred_def_id}",
                cred_def_get,
                allow_head=False,
            ),
            web.get(
                "/anoncreds/credential-definitions",
                cred_defs_get,
                allow_head=False,
            ),
            web.post("/anoncreds/revocation-registry-definition", rev_reg_def_post),
            web.post("/anoncreds/revocation-list", rev_list_post),
            web.put("/anoncreds/registry/{rev_reg_id}/tails-file", upload_tails_file),
            web.put("/anoncreds/registry/{rev_reg_id}/active", set_active_registry),
        ]
    )


def post_process_routes(app: web.Application) -> None:
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": SCHEMAS_TAG_TITLE,
            "description": "AnonCreds schema management",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )
    app._state["swagger_dict"]["tags"].append(
        {
            "name": CRED_DEF_TAG_TITLE,
            "description": "AnonCreds credential definition management",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""DIF Proof Proposal Schema."""

from marshmallow import fields

from ....messaging.models.openapi import OpenAPISchema
from .pres_exch import DIFOptionsSchema, InputDescriptorsSchema


class DIFProofProposalSchema(OpenAPISchema):
    """Schema for DIF Proposal."""

    input_descriptors = fields.List(
        fields.Nested(InputDescriptorsSchema(), required=True), required=False
    )
    options = fields.Nested(DIFOptionsSchema(), required=False)



import json
from unittest import TestCase

from marshmallow import ValidationError

from ..valid import (
    BASE58_SHA256_HASH_VALIDATE,
    BASE64_VALIDATE,
    BASE64URL_NO_PAD_VALIDATE,
    BASE64URL_VALIDATE,
    CREDENTIAL_CONTEXT_VALIDATE,
    CREDENTIAL_TYPE_VALIDATE,
    DID_KEY_VALIDATE,
    DID_POSTURE_VALIDATE,
    DID_WEBVH_VALIDATE,
    ENDPOINT_TYPE_VALIDATE,
    ENDPOINT_VALIDATE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_CRED_REV_ID_VALIDATE,
    INDY_DID_VALIDATE,
    INDY_EXTRA_WQL_VALIDATE,
    INDY_REV_REG_ID_VALIDATE,
    INDY_REV_REG_SIZE_VALIDATE,
    INDY_SCHEMA_ID_VALIDATE,
    INDY_WQL_VALIDATE,
    INT_EPOCH_VALIDATE,
    ISO8601_DATETIME_VALIDATE,
    JWS_HEADER_KID_VALIDATE,
    JWT_VALIDATE,
    MAJOR_MINOR_VERSION_VALIDATE,
    NATURAL_NUM_VALIDATE,
    NUM_STR_NATURAL_VALIDATE,
    NUM_STR_WHOLE_VALIDATE,
    PREDICATE_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
    SHA256_VALIDATE,
    UUID4_VALIDATE,
    WHOLE_NUM_VALIDATE,
)


class TestValid(TestCase):
    def test_epoch(self):
        non_epochs = [-1, 18446744073709551616]
        for non_epoch in non_epochs:
            with self.assertRaises(ValidationError):
                INT_EPOCH_VALIDATE(non_epoch)

        INT_EPOCH_VALIDATE(0)
        INT_EPOCH_VALIDATE(2147483647)
        INT_EPOCH_VALIDATE(18446744073709551615)

    def test_whole(self):
        non_wholes = [-9223372036854775809, 2.3, "Hello", None]
        for non_whole in non_wholes:
            with self.assertRaises(ValidationError):
                WHOLE_NUM_VALIDATE(non_whole)

        WHOLE_NUM_VALIDATE(0)
        WHOLE_NUM_VALIDATE(1)
        WHOLE_NUM_VALIDATE(12345678901234567890)

    def test_natural(self):
        non_naturals = [-9223372036854775809, 2.3, "Hello", None]
        for non_naturals in non_naturals:
            with self.assertRaises(ValidationError):
                NATURAL_NUM_VALIDATE(non_naturals)

        NATURAL_NUM_VALIDATE(1)
        NATURAL_NUM_VALIDATE(12345678901234567890)

    def test_str_whole(self):
        non_str_wholes = ["-9223372036854775809", "Hello", "5.5", "4+3"]
        for non_str_whole in non_str_wholes:
            with self.assertRaises(ValidationError):
                NUM_STR_WHOLE_VALIDATE(non_str_whole)

        NUM_STR_WHOLE_VALIDATE("0")
        NUM_STR_WHOLE_VALIDATE("1")
        NUM_STR_WHOLE_VALIDATE("12345678901234567890")

    def test_str_natural(self):
        non_str_naturals = ["-9223372036854775809", "Hello", "0", "5.5"]
        for non_str_natural in non_str_naturals:
            with self.assertRaises(ValidationError):
                NUM_STR_NATURAL_VALIDATE(non_str_natural)

        NUM_STR_NATURAL_VALIDATE("1")
        NUM_STR_NATURAL_VALIDATE("2")
        NUM_STR_NATURAL_VALIDATE("12345678901234567890")

    def test_indy_rev_reg_size(self):
        non_indy_rr_sizes = [-9223372036854775809, 2.3, "Hello", 0, 3, 32769, None]
        for non_indy_rr_size in non_indy_rr_sizes:
            with self.assertRaises(ValidationError):
                INDY_REV_REG_SIZE_VALIDATE(non_indy_rr_size)

        INDY_REV_REG_SIZE_VALIDATE(4)
        INDY_REV_REG_SIZE_VALIDATE(5)
        INDY_REV_REG_SIZE_VALIDATE(32767)
        INDY_REV_REG_SIZE_VALIDATE(32768)

    def test_indy_did(self):
        non_indy_dids = [
            "Q4zqM7aXqm7gDQkUVLng9I",  # 'I' not a base58 char
            "Q4zqM7aXqm7gDQkUVLng",  # too short
            "Q4zqM7aXqm7gDQkUVLngZZZ",  # too long
            "did:sov:Q4zqM7aXqm7gDQkUVLngZZZ",  # too long
            "did:other:Q4zqM7aXqm7gDQkUVLng9h",  # specifies non-indy DID
        ]
        for non_indy_did in non_indy_dids:
            with self.assertRaises(ValidationError):
                INDY_DID_VALIDATE(non_indy_did)

        INDY_DID_VALIDATE("Q4zqM7aXqm7gDQkUVLng9h")
        INDY_DID_VALIDATE("did:sov:Q4zqM7aXqm7gDQkUVLng9h")

    def test_webvh_did(self):
        valid_webvh_dids = [
            "did:webvh:QmUchSB5f5DJQks9CeyLJjhAy4iKJcYzRyiuYq3sjV13px:example.com",
            "did:webvh:QmZiKXwQVfyZVuvCsuHpQh4arSUpEmeVVRvSfv3uiEycSr:example.com%3A5000",
        ]
        for valid_webvh_did in valid_webvh_dids:
            DID_WEBVH_VALIDATE(valid_webvh_did)

        non_valid_webvh_dids = [
            "did:web:QmUchSB5f5DJQks9CeyLJjhAy4iKJcYzRyiuYq3sjV13px",
            # Did urls are not allowed
            "did:webvh:QmP9VWaTCHcyztDpRj9XSHvZbmYe3m9HZ61KoDtZgWaXVU:example.com%3A5000#z6MkkzY9skorPaoEbCJFKUo7thD8Yb8MBs28aJRopf1TUo9V",
            "did:webvh:QmZiKXwQVfyZVuvCsuHpQh4arSUpEmeVVRvSfv3uiEycSr:example.com%3A5000/whois",
        ]
        for non_valid_webvh_did in non_valid_webvh_dids:
            with self.assertRaises(ValidationError):
                DID_WEBVH_VALIDATE(non_valid_webvh_did)

    def test_indy_raw_public_key(self):
        non_indy_raw_public_keys = [
            "Q4zqM7aXqm7gDQkUVLng9JQ4zqM7aXqm7gDQkUVLng9I",  # 'I' not a base58 char
            "Q4zqM7aXqm7gDQkUVLng",  # too short
            "Q4zqM7aXqm7gDQkUVLngZZZZZZZZZZZZZZZZZZZZZZZZZ",  # too long
        ]
        for non_indy_raw_public_key in non_indy_raw_public_keys:
            with self.assertRaises(ValidationError):
                RAW_ED25519_2018_PUBLIC_KEY_VALIDATE(non_indy_raw_public_key)

        RAW_ED25519_2018_PUBLIC_KEY_VALIDATE(
            "Q4zqM7aXqm7gDQkUVLng9hQ4zqM7aXqm7gDQkUVLng9h"
        )

    def test_jws_header_kid(self):
        non_kids = [
            "http://not-this.one",
            "did:sov:i",  # too short
            "did:key:Q4zqM7aXqm7gDQkUVLng9h"  # missing leading z
            "did:key:zI4zqM7aXqm7gDQkUVLng9h",  # 'I' not a base58 char
        ]
        for non_kid in non_kids:
            with self.assertRaises(ValidationError):
                JWS_HEADER_KID_VALIDATE(non_kid)

        JWS_HEADER_KID_VALIDATE("did:key:zQ4zqM7aXqm7gDQkUVLng9h")
        JWS_HEADER_KID_VALIDATE("did:sov:Q4zqM7aXqm7gDQkUVLng9h#abc-123")
        JWS_HEADER_KID_VALIDATE(
            "did:sov:Q4zqM7aXqm7gDQkUVLng9h?version-time=1234567890#abc-123"
        )
        JWS_HEADER_KID_VALIDATE(
            "did:sov:Q4zqM7aXqm7gDQkUVLng9h?version-time=1234567890&a=b#abc-123"
        )
        JWS_HEADER_KID_VALIDATE(
            "did:sov:Q4zqM7aXqm7gDQkUVLng9h;foo:bar=low;a=b?version-id=1&a=b#abc-123"
        )

    def test_jwt(self):
        non_jwts = [
            "abcde",
            "abcde+.abcde/.abcdef",
            "abcdef==.abcdef==.abcdef",
            "abcdef==..",
        ]

        for non_jwt in non_jwts:
            with self.assertRaises(ValidationError):
                JWT_VALIDATE(non_jwt)

        JWT_VALIDATE("abcdef.abcdef.abcdef")
        JWT_VALIDATE("abcde-.abcde_.abcdef")
        JWT_VALIDATE("abcde-..abcdef")

    def test_did_key(self):
        non_did_keys = [
            "http://not-this.one",
            "did:sov:i",  # wrong preamble
            "did:key:Q4zqM7aXqm7gDQkUVLng9h"  # missing leading z
            "did:key:zI4zqM7aXqm7gDQkUVLng9h",  # 'I' not a base58 char
        ]
        for non_did_key in non_did_keys:
            with self.assertRaises(ValidationError):
                DID_KEY_VALIDATE(non_did_key)

        DID_KEY_VALIDATE("did:key:zQ4zqM7aXqm7gDQkUVLng9h")

    def test_did_posture(self):
        non_did_postures = [
            "not-me",
            None,
            "PUBLIC",
            "Posted",
            "wallet only",
        ]
        for non_did_posture in non_did_postures:
            with self.assertRaises(ValidationError):
                DID_POSTURE_VALIDATE(non_did_posture)

        DID_POSTURE_VALIDATE("public")
        DID_POSTURE_VALIDATE("posted")
        DID_POSTURE_VALIDATE("wallet_only")

    def test_indy_base58_sha256_hash(self):
        non_base58_sha256_hashes = [
            "Q4zqM7aXqm7gDQkUVLng9JQ4zqM7aXqm7gDQkUVLng9I",  # 'I' not a base58 char
            "Q4zqM7aXqm7gDQkUVLng",  # too short
            "Q4zqM7aXqm7gDQkUVLngZZZZZZZZZZZZZZZZZZZZZZZZZ",  # too long
        ]
        for non_base58_sha256_hash in non_base58_sha256_hashes:
            with self.assertRaises(ValidationError):
                BASE58_SHA256_HASH_VALIDATE(non_base58_sha256_hash)

        BASE58_SHA256_HASH_VALIDATE("Q4zqM7aXqm7gDQkUVLng9hQ4zqM7aXqm7gDQkUVLng9h")

    def test_cred_def_id(self):
        non_cred_def_ids = [
            "Q4zqM7aXqm7gDQkUVLng9h:4:CL:18:0",
            "Q4zqM7aXqm7gDQkUVLng9h::CL:18:0",
            "Q4zqM7aXqm7gDQkUVLng9I:3:CL:18:tag",
            "Q4zqM7aXqm7gDQkUVLng9h:3::18:tag",
            "Q4zqM7aXqm7gDQkUVLng9h:3:18:tag",
        ]
        for non_cred_def_id in non_cred_def_ids:
            with self.assertRaises(ValidationError):
                INDY_CRED_DEF_ID_VALIDATE(non_cred_def_id)

        INDY_CRED_DEF_ID_VALIDATE("Q4zqM7aXqm7gDQkUVLng9h:3:CL:18:tag")  # short
        INDY_CRED_DEF_ID_VALIDATE(
            "Q4zqM7aXqm7gDQkUVLng9h:3:CL:Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0:tag"
        )  # long

    def test_rev_reg_id(self):
        non_rev_reg_ids = [
            "WgWxqztrNooG92RXvxSTWv:2:WgWxqztrNooG92RXvxSTWv:3:CL:20:tag:CL_ACCUM:0",
            "WgWxqztrNooG92RXvxSTWI:4:WgWxqztrNooG92RXvxSTWv:3:CL:20:tag:CL_ACCUM:0",
            "WgWxqztrN:4:WgWxqztrNooG92RXvxSTWv:3:CL:20:tag:CL_ACCUM:0",
            "WgWxqztrNooG92RXvxSTWvZ:4:WgWxqztrNooG92RXvxSTWvZ:3:CL:20:tag:CL_XXXXX:0",
            "WgWxqztrNooG92RXvxSTWv::WgWxqztrNooG92RXvxSTWv:3:CL:20:tag:CL_ACCUM:0",
            (
                "WgWxqztrNooG92RXvxSTWv:4:WgWxqztrNooG92RXvxSTWv:3:CL:"
                "Q4zqM7aXqm7gDQkUVLng9h:3:bc-reg:1.0:tag:CL_ACCUM:0"
            ),
            (
                "WgWxqztrNooG92RXvxSTWv:4:WgWxqztrNooG92RXvxSTWv:3:CL:"
                "Q4zqM7aXqm7gDQkUVLng9I:2:bc-reg:1.0:tag:CL_ACCUM:0"
            ),
        ]
        for non_rev_reg_id in non_rev_reg_ids:
            with self.assertRaises(ValidationError):
                INDY_REV_REG_ID_VALIDATE(non_rev_reg_id)

        INDY_REV_REG_ID_VALIDATE(
            "WgWxqztrNooG92RXvxSTWv:4:WgWxqztrNooG92RXvxSTWv:3:CL:20:tag:CL_ACCUM:0",
        )  # short
        INDY_REV_REG_ID_VALIDATE(
            "WgWxqztrNooG92RXvxSTWv:4:WgWxqztrNooG92RXvxSTWv:3:CL:"
            "Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0:tag:CL_ACCUM:0"
        )  # long

    def test_cred_rev_id(self):
        non_cred_rev_ids = ["Wg", "0", "-5", "3.14"]
        for non_cred_rev_id in non_cred_rev_ids:
            with self.assertRaises(ValidationError):
                INDY_CRED_REV_ID_VALIDATE(non_cred_rev_id)

        INDY_CRED_REV_ID_VALIDATE("1")
        INDY_CRED_REV_ID_VALIDATE("99999999")

    def test_version(self):
        non_versions = ["-1", "", "3_5", "3.5a"]
        for non_version in non_versions:
            with self.assertRaises(ValidationError):
                MAJOR_MINOR_VERSION_VALIDATE(non_version)

        MAJOR_MINOR_VERSION_VALIDATE("1.0")
        MAJOR_MINOR_VERSION_VALIDATE(".05")
        MAJOR_MINOR_VERSION_VALIDATE("1.2.3")
        MAJOR_MINOR_VERSION_VALIDATE("..")  # perverse but technically OK

    def test_schema_id(self):
        non_schema_ids = [
            "Q4zqM7aXqm7gDQkUVLng9h:3:bc-reg:1.0",
            "Q4zqM7aXqm7gDQkUVLng9h::bc-reg:1.0",
            "Q4zqM7aXqm7gDQkUVLng9h:bc-reg:1.0",
            "Q4zqM7aXqm7gDQkUVLng9h:2:1.0",
            "Q4zqM7aXqm7gDQkUVLng9h:2::1.0",
            "Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:",
            "Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0a",
            "Q4zqM7aXqm7gDQkUVLng9I:2:bc-reg:1.0",  # I is not in base58
        ]
        for non_schema_id in non_schema_ids:
            with self.assertRaises(ValidationError):
                INDY_SCHEMA_ID_VALIDATE(non_schema_id)

        INDY_SCHEMA_ID_VALIDATE("Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0")

    def test_predicate(self):
        non_predicates = [">>", "", " >= ", "<<<=", "==", "=", "!="]
        for non_predicate in non_predicates:
            with self.assertRaises(ValidationError):
                PREDICATE_VALIDATE(non_predicate)

        PREDICATE_VALIDATE("<")
        PREDICATE_VALIDATE("<=")
        PREDICATE_VALIDATE(">=")
        PREDICATE_VALIDATE(">")

    def test_indy_date(self):
        non_datetimes = [
            "nope",
            "2020-01-01",
            "2020-01-01:00:00:00Z",
            "2020.01.01 00:00:00Z",
            "2020-01-01T00:00.123456+00:00",
            "2020-01-01T00:00:00.123456+0:00",
        ]
        for non_datetime in non_datetimes:
            with self.assertRaises(ValidationError):
                ISO8601_DATETIME_VALIDATE(non_datetime)

        ISO8601_DATETIME_VALIDATE("2020-01-01 00:00:00Z")
        ISO8601_DATETIME_VALIDATE("2020-01-01T00:00:00Z")
        ISO8601_DATETIME_VALIDATE("2020-01-01T00:00:00")
        ISO8601_DATETIME_VALIDATE("2020-01-01 00:00:00")
        ISO8601_DATETIME_VALIDATE("2020-01-01 00:00:00+00:00")
        ISO8601_DATETIME_VALIDATE("2020-01-01 00:00:00-00:00")
        ISO8601_DATETIME_VALIDATE("2020-01-01 00:00-00:00")
        ISO8601_DATETIME_VALIDATE("2020-01-01 00:00:00.1-00:00")
        ISO8601_DATETIME_VALIDATE("2020-01-01 00:00:00.123456-00:00")

    def test_indy_wql(self):
        non_wqls = [
            "nope",
            "[a, b, c]",
            "{1, 2, 3}",
            set(),
            '"Hello World"',
            None,
            "null",
            "true",
            False,
        ]
        for non_wql in non_wqls:
            with self.assertRaises(ValidationError):
                INDY_WQL_VALIDATE(non_wql)

        INDY_WQL_VALIDATE(json.dumps({}))
        INDY_WQL_VALIDATE(json.dumps({"a": "1234"}))
        INDY_WQL_VALIDATE(json.dumps({"a": "1234", "b": {"$not": "0"}}))
        INDY_WQL_VALIDATE(json.dumps({"$or": {"a": "1234", "b": "0"}}))

    def test_indy_extra_wql(self):
        non_xwqls = [
            "nope",
            "[a, b, c]",
            "{1, 2, 3}",
            set(),
            '"Hello World"',
            None,
            "null",
            "true",
            False,
            "{}",
            '{"no": "referent"}',
            '{"no": "referent", "another": "non-referent"}',
            '{"uuid": {"too many: "braces"}}}',
        ]
        for non_xwql in non_xwqls:
            with self.assertRaises(ValidationError):
                INDY_EXTRA_WQL_VALIDATE(non_xwql)

        INDY_EXTRA_WQL_VALIDATE(json.dumps({"uuid0": {"name::ident::marker": "1"}}))
        INDY_EXTRA_WQL_VALIDATE(
            json.dumps(
                {
                    "uuid0": {"attr::ident::marker": "1"},
                    "uuid1": {"attr::member::value": "655321"},
                    "uuid2": {"attr::code::value": {"$in": ["abc", "def", "ghi"]}},
                    "uuid3": {"attr::score::value": {"$neq": "0"}},
                }
            )
        )

    def test_base64(self):
        non_base64s = [
            "####",
            "abcde===",
            "abcd====",
            "=abcd123",
            "=abcd123=",
        ]

        for non_base64 in non_base64s:
            with self.assertRaises(ValidationError):
                BASE64_VALIDATE(non_base64)
            with self.assertRaises(ValidationError):
                BASE64URL_VALIDATE(non_base64)

        BASE64_VALIDATE("")
        BASE64_VALIDATE("abcd123")  # some specs like JWT insist on no padding
        BASE64_VALIDATE("abcde")
        BASE64_VALIDATE("UG90YX+v")
        BASE64_VALIDATE("UG90Y+/=")
        BASE64_VALIDATE("UG90YX==")
        with self.assertRaises(ValidationError):
            BASE64_VALIDATE("UG90YX-v")

        BASE64URL_VALIDATE("")
        BASE64URL_VALIDATE("abcd123")  # some specs like JWT insist on no padding
        BASE64URL_VALIDATE("abcde")
        BASE64URL_VALIDATE("UG90YX-v")
        BASE64URL_VALIDATE("UG90Y-_=")
        BASE64URL_VALIDATE("UG90YX==")
        with self.assertRaises(ValidationError):
            BASE64URL_VALIDATE("UG90YX+v")  # '+' is not a base64url char

        non_base64_no_pads = ["####", "abcde=", "ab=cde"]
        for non_base64_no_pad in non_base64_no_pads:
            with self.assertRaises(ValidationError):
                BASE64URL_NO_PAD_VALIDATE(non_base64_no_pad)

        BASE64URL_NO_PAD_VALIDATE("")
        BASE64URL_NO_PAD_VALIDATE("abcd123")
        BASE64URL_NO_PAD_VALIDATE("abcde")
        BASE64URL_NO_PAD_VALIDATE("UG90YX-v")
        BASE64URL_NO_PAD_VALIDATE("UG90Y-_")
        BASE64URL_NO_PAD_VALIDATE("UG90YX")
        with self.assertRaises(ValidationError):
            BASE64URL_NO_PAD_VALIDATE("UG90YX+v")  # '+' is not a base64url char

    def test_sha256(self):
        non_sha256s = [
            "####",
            "abcd123",
            "________________________________________________________________",
            "gggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggg",
        ]
        for non_sha256 in non_sha256s:
            with self.assertRaises(ValidationError):
                SHA256_VALIDATE(non_sha256)

        SHA256_VALIDATE(
            "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
        )
        SHA256_VALIDATE(
            "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        )

    def test_uuid4(self):
        non_uuid4s = [
            "123",
            "",
            "----",
            "3fa85f6-5717-4562-b3fc-2c963f66afa6",  # short a hex digit
            "3fa85f645-5717-4562-b3fc-2c963f66afa6",  # extra hex digit
            "3fa85f64-5717-f562-b3fc-2c963f66afa6",  # 13th hex digit is not 4
        ]
        for non_uuid4 in non_uuid4s:
            with self.assertRaises(ValidationError):
                UUID4_VALIDATE(non_uuid4)

        UUID4_VALIDATE("3fa85f64-5717-4562-b3fc-2c963f66afa6")
        UUID4_VALIDATE("3FA85F64-5717-4562-B3FC-2C963F66AFA6")  # upper case OK

    def test_endpoint(self):
        non_endpoints = [
            "123",
            "",
            "/path/only",
            "https://1.2.3.4?query=true&url=false",
            "no-proto:8080/my/path",
            "smtp:8080/my/path#fragment",
        ]

        for non_endpoint in non_endpoints:
            with self.assertRaises(ValidationError):
                ENDPOINT_VALIDATE(non_endpoint)

        ENDPOINT_VALIDATE("http://github.com")
        ENDPOINT_VALIDATE("https://localhost:8080")
        ENDPOINT_VALIDATE("newproto://myhost.ca:8080/path")
        ENDPOINT_VALIDATE("ftp://10.10.100.90:8021")
        ENDPOINT_VALIDATE("zzzp://someplace.ca:9999/path")

    def test_endpoint_type(self):
        non_endpoint_types = [
            "123",
            "endpoint",
            "end point",
            "end-point",
            "profile",
            "linked_domains",
            None,
        ]

        for non_endpoint_type in non_endpoint_types:
            with self.assertRaises(ValidationError):
                ENDPOINT_TYPE_VALIDATE(non_endpoint_type)

        ENDPOINT_TYPE_VALIDATE("Endpoint")
        ENDPOINT_TYPE_VALIDATE("Profile")
        ENDPOINT_TYPE_VALIDATE("LinkedDomains")

    def test_credential_type(self):
        with self.assertRaises(ValidationError):
            CREDENTIAL_TYPE_VALIDATE([])

        with self.assertRaises(ValidationError):
            CREDENTIAL_TYPE_VALIDATE(["WrongType", "AnotherWrongType"])

        CREDENTIAL_TYPE_VALIDATE(["VerifiableCredential"])
        CREDENTIAL_TYPE_VALIDATE(["VerifiableCredential", "AnotherType"])
        CREDENTIAL_TYPE_VALIDATE(["SomeType", "AnotherType", "VerifiableCredential"])

    def test_credential_context(self):
        with self.assertRaises(ValidationError):
            CREDENTIAL_CONTEXT_VALIDATE([])

        with self.assertRaises(ValidationError):
            CREDENTIAL_CONTEXT_VALIDATE([{}, "https://www.w3.org/2018/credentials/v1"])

        CREDENTIAL_CONTEXT_VALIDATE(["https://www.w3.org/2018/credentials/v1"])
        CREDENTIAL_CONTEXT_VALIDATE(
            ["https://www.w3.org/2018/credentials/v1", "https://some-other-context.com"]
        )



"""Represents an explicit RFC 15 ack message, adopted into present-proof protocol."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import NOTIF_10_ACK, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.ack_handler.V10AckHandler"


class V10Ack(AgentMessage):
    """Base class representing an explicit ack message for no specific protocol."""

    class Meta:
        """V10Ack metadata."""

        handler_class = HANDLER_CLASS
        message_type = NOTIF_10_ACK
        schema_class = "V10AckSchema"

    def __init__(self, status: Optional[str] = None, **kwargs):
        """Initialize an explicit ack message instance.

        Args:
            status: Status (default OK)
            kwargs: Additional message keyword arguments

        """
        super().__init__(**kwargs)
        self.status = status or "OK"


class V10AckSchema(AgentMessageSchema):
    """Schema for V10Ack class."""

    class Meta:
        """V10Ack schema metadata."""

        model_class = V10Ack
        unknown = EXCLUDE

    status = fields.Str(
        required=True, metadata={"description": "Status", "example": "OK"}
    )



"""A simple event bus."""

import asyncio
import logging
from contextlib import contextmanager
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterator,
    List,
    Match,
    NamedTuple,
    Optional,
    Pattern,
    Tuple,
)

if TYPE_CHECKING:  # To avoid circular import error
    from .profile import Profile

LOGGER = logging.getLogger(__name__)


class Event:
    """A simple event object."""

    def __init__(self, topic: str, payload: Optional[Any] = None):
        """Create a new event."""
        self._topic = topic
        self._payload = payload

    @property
    def topic(self):
        """Return this event's topic."""
        return self._topic

    @property
    def payload(self):
        """Return this event's payload."""
        return self._payload

    def __eq__(self, other):
        """Test equality."""
        if not isinstance(other, Event):
            return False
        return self._topic == other._topic and self._payload == other._payload

    def __repr__(self):
        """Return debug representation."""
        return "<Event topic={}, payload={}>".format(self._topic, self._payload)

    def with_metadata(self, metadata: "EventMetadata") -> "EventWithMetadata":
        """Annotate event with metadata and return EventWithMetadata object."""
        return EventWithMetadata(self.topic, self.payload, metadata)


class EventMetadata(NamedTuple):
    """Metadata passed alongside events to add context."""

    pattern: Pattern
    match: Match[str]


class EventWithMetadata(Event):
    """Event with metadata passed alongside events to add context."""

    def __init__(self, topic: str, payload: Any, metadata: EventMetadata):
        """Initialize event metadata."""
        super().__init__(topic, payload)
        self._metadata = metadata

    @property
    def metadata(self) -> EventMetadata:
        """Return metadata."""
        return self._metadata


class EventBus:
    """A simple event bus implementation."""

    def __init__(self):
        """Initialize Event Bus."""
        self.topic_patterns_to_subscribers: Dict[Pattern, List[Callable]] = {}

    async def notify(self, profile: "Profile", event: Event):
        """Notify subscribers of event.

        Args:
            profile (Profile): context of the event
            event (Event): event to emit

        """
        # TODO don't block notifier until subscribers have all been called?
        # TODO trigger each processor but don't await?
        # TODO log errors but otherwise ignore?

        LOGGER.debug("Notifying subscribers: %s", event)

        partials = []
        for pattern, subscribers in self.topic_patterns_to_subscribers.items():
            match = pattern.match(event.topic)

            if not match:
                continue

            for subscriber in subscribers:
                partials.append(
                    partial(
                        subscriber,
                        profile,
                        event.with_metadata(EventMetadata(pattern, match)),
                    )
                )

        for processor in partials:
            try:
                await processor()
            except Exception:
                LOGGER.exception("Error occurred while processing event")

    def subscribe(self, pattern: Pattern, processor: Callable):
        """Subscribe to an event.

        Args:
            pattern (Pattern): compiled regular expression for matching topics
            processor (Callable): async callable accepting profile and event

        """
        LOGGER.debug("Subscribed: topic %s, processor %s", pattern, processor)
        if pattern not in self.topic_patterns_to_subscribers:
            self.topic_patterns_to_subscribers[pattern] = []
        self.topic_patterns_to_subscribers[pattern].append(processor)

    def unsubscribe(self, pattern: Pattern, processor: Callable):
        """Unsubscribe from an event.

        This method is idempotent. Repeated calls to unsubscribe will not
        result in errors.

        Args:
            pattern (Pattern): regular expression used to subscribe the processor
            processor (Callable): processor to unsubscribe

        """
        if pattern in self.topic_patterns_to_subscribers:
            try:
                index = self.topic_patterns_to_subscribers[pattern].index(processor)
            except ValueError:
                return
            del self.topic_patterns_to_subscribers[pattern][index]
            if not self.topic_patterns_to_subscribers[pattern]:
                del self.topic_patterns_to_subscribers[pattern]
            LOGGER.debug("Unsubscribed: topic %s, processor %s", pattern, processor)

    @contextmanager
    def wait_for_event(
        self,
        waiting_profile: "Profile",
        pattern: Pattern,
        cond: Optional[Callable[[Event], bool]] = None,
    ) -> Iterator[Awaitable[Event]]:
        """Capture an event and retrieve its value."""
        future = asyncio.get_event_loop().create_future()

        async def _handle_single_event(profile, event):
            """Handle the single event."""
            LOGGER.debug(
                "wait_for_event event listener with event %s and profile %s",
                event,
                profile,
            )
            if cond is not None and not cond(event):
                return

            if waiting_profile == profile:
                future.set_result(event)
                self.unsubscribe(pattern, _handle_single_event)

        self.subscribe(pattern, _handle_single_event)

        yield future

        if not future.done():
            future.cancel()


class MockEventBus(EventBus):
    """A mock EventBus for testing."""

    def __init__(self):
        """Initialize MockEventBus."""
        super().__init__()
        self.events: List[Tuple[Profile, Event]] = []

    async def notify(self, profile: "Profile", event: Event):
        """Append the event to MockEventBus.events."""
        self.events.append((profile, event))



"""Class to manage transactions."""

import json
import logging
from asyncio import shield
from typing import Optional

from uuid_utils import uuid4

from ....anoncreds.issuer import AnonCredsIssuer
from ....anoncreds.revocation import AnonCredsRevocation
from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from ....core.profile import Profile
from ....indy.issuer import IndyIssuerError
from ....ledger.base import BaseLedger
from ....ledger.error import LedgerError
from ....messaging.credential_definitions.util import notify_cred_def_event
from ....messaging.schemas.util import notify_schema_event
from ....revocation.util import (
    notify_revocation_entry_endorsed_event,
    notify_revocation_reg_endorsed_event,
)
from ....storage.error import StorageError, StorageNotFoundError
from ....wallet.base import BaseWallet
from ....wallet.util import notify_endorse_did_attrib_event, notify_endorse_did_event
from .messages.cancel_transaction import CancelTransaction
from .messages.endorsed_transaction_response import EndorsedTransactionResponse
from .messages.refused_transaction_response import RefusedTransactionResponse
from .messages.transaction_acknowledgement import TransactionAcknowledgement
from .messages.transaction_job_to_send import TransactionJobToSend
from .messages.transaction_request import TransactionRequest
from .messages.transaction_resend import TransactionResend
from .models.transaction_record import TransactionRecord
from .transaction_jobs import TransactionJob


class TransactionManagerError(BaseError):
    """Transaction error."""


class TransactionManager:
    """Class for managing transactions."""

    def __init__(self, profile: Profile):
        """Initialize a TransactionManager.

        Args:
            profile: The profile instance for this transaction manager
        """
        self._profile = profile
        self._logger = logging.getLogger(__name__)

    @property
    def profile(self) -> Profile:
        """Accessor for the current Profile.

        Returns:
            The Profile for this transaction manager

        """
        return self._profile

    async def create_record(
        self, messages_attach: str, connection_id: str, meta_data: Optional[dict] = None
    ):
        """Create a new Transaction Record.

        Args:
            messages_attach: messages to attach, JSON-dumped
            connection_id: The connection_id of the ConnRecord between author and endorser
            meta_data: The meta_data to attach to the transaction record

        Returns:
            The transaction Record

        """

        messages_attach_dict = {
            "@id": str(uuid4()),
            "mime-type": "application/json",
            "data": {"json": messages_attach},
        }

        transaction = TransactionRecord()

        formats = {
            "attach_id": messages_attach_dict["@id"],
            "format": TransactionRecord.FORMAT_VERSION,
        }
        transaction.formats.clear()
        transaction.formats.append(formats)

        transaction.messages_attach.clear()
        transaction.messages_attach.append(messages_attach_dict)

        if meta_data:
            transaction.meta_data = meta_data

        transaction.state = TransactionRecord.STATE_TRANSACTION_CREATED
        transaction.connection_id = connection_id

        async with self._profile.session() as session:
            await transaction.save(session, reason="Created a Transaction Record")

        return transaction

    # todo - implementing changes for writing final transaction to the ledger
    # (For Sign Transaction Protocol)
    async def create_request(
        self,
        transaction: TransactionRecord,
        signature: Optional[str] = None,
        signed_request: Optional[dict] = None,
        expires_time: Optional[str] = None,
        endorser_write_txn: bool = False,
        author_goal_code: Optional[str] = None,
        signer_goal_code: Optional[str] = None,
    ):
        """Create a new Transaction Request.

        Args:
            transaction (TransactionRecord): The transaction from which the request is
                created.
            signature (str, optional): The signature for the transaction.
                Defaults to None.
            signed_request (dict, optional): The signed request for the transaction.
                Defaults to None.
            expires_time (str, optional): The time till which the endorser should endorse
                the transaction. Defaults to None.
            endorser_write_txn (bool, optional): Flag indicating if the endorser should
                write the transaction. Defaults to False.
            author_goal_code (str, optional): The goal code for the author.
                Defaults to None.
            signer_goal_code (str, optional): The goal code for the signer.
                Defaults to None.

        Returns:
            Tuple[TransactionRecord, TransactionRequest]: The transaction record and
                transaction request.

        Raises:
            TransactionManagerError: If the transaction record is not in the
                'STATE_TRANSACTION_CREATED' state.

        """

        if transaction.state != TransactionRecord.STATE_TRANSACTION_CREATED:
            raise TransactionManagerError(
                f"Cannot create a request for transaction record"
                f" in state: {transaction.state}"
            )

        transaction._type = TransactionRecord.SIGNATURE_REQUEST
        signature_request = {
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signer_goal_code": (
                signer_goal_code
                if signer_goal_code
                else TransactionRecord.ENDORSE_TRANSACTION
            ),
            "author_goal_code": (
                author_goal_code
                if author_goal_code
                else TransactionRecord.WRITE_TRANSACTION
            ),
        }
        transaction.signature_request.clear()
        transaction.signature_request.append(signature_request)

        transaction.state = TransactionRecord.STATE_REQUEST_SENT

        timing = {"expires_time": expires_time}
        transaction.timing = timing
        transaction.endorser_write_txn = endorser_write_txn

        async with self._profile.session() as session:
            await transaction.save(session, reason="Created an endorsement request")

        transaction_request = TransactionRequest(
            transaction_id=transaction._id,
            signature_request=transaction.signature_request[0],
            timing=transaction.timing,
            messages_attach=transaction.messages_attach[0],
            endorser_write_txn=endorser_write_txn,
        )

        return transaction, transaction_request

    async def receive_request(self, request: TransactionRequest, connection_id: str):
        """Receive a Transaction request.

        Args:
            request: A Transaction Request
            connection_id: The connection id related to this transaction record
        """

        transaction = TransactionRecord()

        transaction._type = TransactionRecord.SIGNATURE_REQUEST
        transaction.signature_request.clear()
        transaction.signature_request.append(request.signature_request)
        transaction.timing = request.timing

        format = {
            "attach_id": request.messages_attach["@id"],
            "format": TransactionRecord.FORMAT_VERSION,
        }
        transaction.formats.clear()
        transaction.formats.append(format)

        transaction.messages_attach.clear()
        transaction.messages_attach.append(request.messages_attach)
        transaction.thread_id = request.transaction_id
        transaction.connection_id = connection_id
        transaction.state = TransactionRecord.STATE_REQUEST_RECEIVED
        transaction.endorser_write_txn = request.endorser_write_txn

        async with self._profile.session() as session:
            await transaction.save(session, reason="Received an endorsement request")

        return transaction

    async def create_endorse_response(
        self,
        transaction: TransactionRecord,
        state: str,
        use_endorser_did: Optional[str] = None,
    ):
        """Create a response to endorse a transaction.

        Args:
            transaction: The transaction record which would be endorsed.
            state: The state of the transaction record
            use_endorser_did: The public did of the endorser

        Returns:
            The updated transaction and an endorsed response

        """

        if transaction.state not in (
            TransactionRecord.STATE_REQUEST_RECEIVED,
            TransactionRecord.STATE_TRANSACTION_RESENT_RECEIVED,
        ):
            raise TransactionManagerError(
                f"Cannot endorse transaction for transaction record"
                f" in state: {transaction.state}"
            )

        transaction._type = TransactionRecord.SIGNATURE_RESPONSE
        transaction_json = transaction.messages_attach[0]["data"]["json"]
        ledger_response = {}

        async with self._profile.session() as session:
            wallet: BaseWallet = session.inject_or(BaseWallet)
            if not wallet:
                raise StorageError("No wallet available")
            endorser_did_info = None
            override_did = (
                use_endorser_did
                if use_endorser_did
                else session.context.settings.get_value(
                    "endorser.endorser_endorse_with_did"
                )
            )
            if override_did:
                endorser_did_info = await wallet.get_local_did(override_did)
            else:
                endorser_did_info = await wallet.get_public_did()
            if not endorser_did_info:
                raise StorageError(
                    "Transaction cannot be endorsed as there is no Public DID in wallet "
                    "or Endorser DID specified"
                )
            endorser_did = endorser_did_info.did
            endorser_verkey = endorser_did_info.verkey

        ledger = self._profile.context.inject_or(BaseLedger)
        if not ledger:
            reason = "No ledger available"
            if not self._profile.context.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise LedgerError(reason=reason)

        async with ledger:
            # check our goal code!
            txn_goal_code = (
                transaction.signature_request[0]["signer_goal_code"]
                if transaction.signature_request
                and "signer_goal_code" in transaction.signature_request[0]
                else TransactionRecord.ENDORSE_TRANSACTION
            )
            if txn_goal_code == TransactionRecord.ENDORSE_TRANSACTION:
                endorsed_msg = await shield(
                    ledger.txn_endorse(transaction_json, endorse_did=endorser_did_info)
                )
            elif txn_goal_code == TransactionRecord.WRITE_DID_TRANSACTION:
                # get DID info from transaction.meta_data
                meta_data = json.loads(transaction_json)
                (success, txn) = await shield(
                    ledger.register_nym(
                        meta_data["did"],
                        meta_data["verkey"],
                        meta_data["alias"],
                        meta_data["role"],
                    )
                )
                # we don't have an endorsed transaction so just return did meta-data
                ledger_response = {
                    "result": {"txn": {"type": "1", "data": {"dest": meta_data["did"]}}},
                    "meta_data": meta_data,
                }
                endorsed_msg = json.dumps(ledger_response)
            else:
                raise TransactionManagerError(
                    f"Invalid goal code for transaction record: {txn_goal_code}"
                )

        # need to return the endorsed msg or else the ledger will reject the
        # eventual transaction write
        transaction.messages_attach[0]["data"]["json"] = endorsed_msg

        signature_response = {
            "message_id": transaction.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": txn_goal_code,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signature": {endorser_did: endorsed_msg or endorser_verkey},
        }

        transaction.signature_response.clear()
        transaction.signature_response.append(signature_response)

        transaction.state = state

        async with self._profile.session() as session:
            await transaction.save(session, reason="Created an endorsed response")

        if (
            transaction.endorser_write_txn
            and txn_goal_code == TransactionRecord.ENDORSE_TRANSACTION
        ):
            # no longer supported - if the author asks the endorser to write
            # the transaction, raise an error
            raise TransactionManagerError(
                "Operation not supported, endorser cannot write the ledger transaction"
            )

        endorsed_transaction_response = EndorsedTransactionResponse(
            transaction_id=transaction.thread_id,
            thread_id=transaction._id,
            signature_response=signature_response,
            state=state,
            endorser_did=endorser_did,
            ledger_response=ledger_response,
        )

        return transaction, endorsed_transaction_response

    async def receive_endorse_response(self, response: EndorsedTransactionResponse):
        """Update the transaction record with the endorsed response.

        Args:
            response: The Endorsed Transaction Response
        """

        async with self._profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_id(
                session, response.transaction_id
            )

        transaction._type = TransactionRecord.SIGNATURE_RESPONSE
        transaction.state = response.state

        transaction.signature_response.clear()
        transaction.signature_response.append(response.signature_response)

        transaction.thread_id = response.thread_id

        # the returned signature is actually the endorsed ledger transaction
        endorser_did = response.endorser_did
        transaction.messages_attach[0]["data"]["json"] = response.signature_response[
            "signature"
        ][endorser_did]

        async with self._profile.session() as session:
            await transaction.save(session, reason="Received an endorsed response")

        # this scenario is where the author has asked the endorser to write the ledger
        # we are not supporting endorser ledger writes ...
        if transaction.endorser_write_txn:
            raise TransactionManagerError(
                "Endorser ledger writes are no longer supported"
            )

        return transaction

    async def complete_transaction(
        self, transaction: TransactionRecord, endorser: bool = False
    ):
        """Complete a transaction.

        This is the final state where the received ledger transaction
        is written to the ledger.

        Args:
            transaction: The transaction record which would be completed
            endorser: True if the endorser is writing the ledger transaction

        Returns:
            The updated transaction

        """

        ledger_transaction = transaction.messages_attach[0]["data"]["json"]

        # check our goal code!
        txn_goal_code = (
            transaction.signature_request[0]["signer_goal_code"]
            if transaction.signature_request
            and "signer_goal_code" in transaction.signature_request[0]
            else TransactionRecord.ENDORSE_TRANSACTION
        )

        # if we are the author, we need to write the endorsed ledger transaction ...
        # ... EXCEPT for DID transactions, which the endorser will write
        if (not endorser) and (txn_goal_code != TransactionRecord.WRITE_DID_TRANSACTION):
            ledger = self.profile.inject(BaseLedger)
            if not ledger:
                raise TransactionManagerError("No ledger available")
            if (
                self._profile.context.settings.get_value("wallet.type")
                == "askar-anoncreds"
            ):
                from acapy_agent.anoncreds.default.legacy_indy.registry import (
                    LegacyIndyRegistry,
                )

                legacy_indy_registry = LegacyIndyRegistry()
                ledger_response_json = await shield(
                    legacy_indy_registry.txn_submit(
                        ledger, ledger_transaction, sign=False, taa_accept=False
                    )
                )
            else:
                async with ledger:
                    try:
                        ledger_response_json = await shield(
                            ledger.txn_submit(
                                ledger_transaction, sign=False, taa_accept=False
                            )
                        )
                    except (IndyIssuerError, LedgerError) as err:
                        raise TransactionManagerError(err.roll_up) from err

            ledger_response = json.loads(ledger_response_json)

        else:
            ledger_response = ledger_transaction

        transaction.state = TransactionRecord.STATE_TRANSACTION_ACKED

        async with self._profile.session() as session:
            await transaction.save(session, reason="Completed transaction")

        # this scenario is where the endorser is writing the transaction
        # shouldn't get to this point, but check and raise an error just in case
        if endorser and transaction.endorser_write_txn:
            raise TransactionManagerError(
                "Operation not supported, endorser cannot write the ledger transaction"
            )

        connection_id = transaction.connection_id
        async with self._profile.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            jobs = await connection_record.metadata_get(session, "transaction_jobs")
        if not jobs:
            raise TransactionManagerError(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        if "transaction_my_job" not in jobs.keys():
            raise TransactionManagerError(
                'The "transaction_my_job" is not set in "transaction_jobs"'
                " in connection metadata for this connection record"
            )
        if jobs["transaction_my_job"] == TransactionJob.TRANSACTION_AUTHOR.name:
            # the author write the endorsed transaction to the ledger
            await self.endorsed_txn_post_processing(
                transaction, ledger_response, connection_record
            )
            transaction_acknowledgement_message = TransactionAcknowledgement(
                thread_id=transaction._id
            )

        elif jobs["transaction_my_job"] == TransactionJob.TRANSACTION_ENDORSER.name:
            transaction_acknowledgement_message = TransactionAcknowledgement(
                thread_id=transaction._id, ledger_response=ledger_response
            )

        return transaction, transaction_acknowledgement_message

    async def receive_transaction_acknowledgement(
        self, response: TransactionAcknowledgement, connection_id: str
    ):
        """Update the transaction record after receiving the transaction acknowledgement.

        Args:
            response: The transaction acknowledgement
            connection_id: The connection_id related to this Transaction Record
        """

        async with self._profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_connection_and_thread(
                session, connection_id, response.thread_id
            )

        if transaction.state != TransactionRecord.STATE_TRANSACTION_ENDORSED:
            raise TransactionManagerError(
                "Only an endorsed transaction can be written to the ledger."
            )

        transaction.state = TransactionRecord.STATE_TRANSACTION_ACKED
        async with self._profile.session() as session:
            await transaction.save(session, reason="Received a transaction ack")

        connection_id = transaction.connection_id

        try:
            async with self._profile.session() as session:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
                jobs = await connection_record.metadata_get(session, "transaction_jobs")
        except StorageNotFoundError as err:
            raise TransactionManagerError(err.roll_up) from err

        is_auto_endorser = self._profile.settings.get(
            "endorser.endorser"
        ) and self._profile.settings.get("endorser.auto_endorse")

        if is_auto_endorser:
            return transaction

        if not jobs:
            raise TransactionManagerError(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        if "transaction_my_job" not in jobs.keys():
            raise TransactionManagerError(
                'The "transaction_my_job" is not set in "transaction_jobs"'
                " in connection metadata for this connection record"
            )
        if jobs["transaction_my_job"] == TransactionJob.TRANSACTION_AUTHOR.name:
            # store the related non-secrets record in our wallet
            await self.endorsed_txn_post_processing(
                transaction, response.ledger_response, connection_record
            )

        return transaction

    async def create_refuse_response(
        self, transaction: TransactionRecord, state: str, refuser_did: str
    ):
        """Create a response to refuse a transaction.

        Args:
            transaction: The transaction record which would be refused
            state: The state of the transaction record
            refuser_did: The public did of the refuser

        Returns:
            The updated transaction and the refused response

        """

        if transaction.state not in (
            TransactionRecord.STATE_REQUEST_RECEIVED,
            TransactionRecord.STATE_TRANSACTION_RESENT_RECEIVED,
        ):
            raise TransactionManagerError(
                f"Cannot refuse transaction for transaction record"
                f" in state: {transaction.state}"
            )

        transaction._type = TransactionRecord.SIGNATURE_RESPONSE

        signature_response = {
            "message_id": transaction.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.REFUSE_TRANSACTION,
        }
        transaction.signature_response.clear()
        transaction.signature_response.append(signature_response)

        transaction.state = state

        async with self._profile.session() as session:
            await transaction.save(session, reason="Created a refused response")

        refused_transaction_response = RefusedTransactionResponse(
            transaction_id=transaction.thread_id,
            thread_id=transaction._id,
            signature_response=signature_response,
            state=state,
            endorser_did=refuser_did,
        )

        return transaction, refused_transaction_response

    async def receive_refuse_response(self, response: RefusedTransactionResponse):
        """Update the transaction record with a refused response.

        Args:
            response: The refused transaction response
        """

        async with self._profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_id(
                session, response.transaction_id
            )

        transaction._type = TransactionRecord.SIGNATURE_RESPONSE
        transaction.state = response.state

        transaction.signature_response.clear()
        transaction.signature_response.append(response.signature_response)
        transaction.thread_id = response.thread_id

        async with self._profile.session() as session:
            await transaction.save(session, reason="Received a refused response")

        return transaction

    async def cancel_transaction(self, transaction: TransactionRecord, state: str):
        """Cancel a Transaction Request.

        Args:
            transaction: The transaction record which would be cancelled
            state: The state of the transaction record

        Returns:
            The updated transaction and the cancelled transaction response

        """

        if transaction.state not in (
            TransactionRecord.STATE_REQUEST_SENT,
            TransactionRecord.STATE_TRANSACTION_RESENT,
        ):
            raise TransactionManagerError(
                f"Cannot cancel transaction as transaction is"
                f" in state: {transaction.state}"
            )

        transaction.state = state
        async with self._profile.session() as session:
            await transaction.save(session, reason="Cancelled the transaction")

        cancelled_transaction_response = CancelTransaction(
            state=state, thread_id=transaction._id
        )

        return transaction, cancelled_transaction_response

    async def receive_cancel_transaction(
        self, response: CancelTransaction, connection_id: str
    ):
        """Update the transaction record to cancel a transaction request.

        Args:
            response: The cancel transaction response
            connection_id: The connection_id related to this Transaction Record
        """

        async with self._profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_connection_and_thread(
                session, connection_id, response.thread_id
            )

        transaction.state = response.state
        async with self._profile.session() as session:
            await transaction.save(session, reason="Received a cancel request")

        return transaction

    async def transaction_resend(self, transaction: TransactionRecord, state: str):
        """Resend a transaction request.

        Args:
            transaction: The transaction record which needs to be resend
            state: the state of the transaction record

        Returns:
            The updated transaction and the resend response

        """

        if transaction.state not in (
            TransactionRecord.STATE_TRANSACTION_REFUSED,
            TransactionRecord.STATE_TRANSACTION_CANCELLED,
        ):
            raise TransactionManagerError(
                f"Cannot resend transaction as transaction is"
                f" in state: {transaction.state}"
            )

        transaction.state = state
        async with self._profile.session() as session:
            await transaction.save(session, reason="Resends the transaction request")

        resend_transaction_response = TransactionResend(
            state=TransactionRecord.STATE_TRANSACTION_RESENT_RECEIVED,
            thread_id=transaction._id,
        )

        return transaction, resend_transaction_response

    async def receive_transaction_resend(
        self, response: TransactionResend, connection_id: str
    ):
        """Update the transaction with a resend request.

        Args:
            response: The Resend transaction response
            connection_id: The connection_id related to this Transaction Record
        """

        async with self._profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_connection_and_thread(
                session, connection_id, response.thread_id
            )

        transaction.state = response.state
        async with self._profile.session() as session:
            await transaction.save(session, reason="Receives a transaction request")

        return transaction

    async def set_transaction_my_job(self, record: ConnRecord, transaction_my_job: str):
        """Set transaction_my_job.

        Args:
            record: The connection record in which to set transaction jobs
            transaction_my_job: My transaction job

        Returns:
            The transaction job that is send to other agent

        """

        async with self._profile.session() as session:
            value = await record.metadata_get(session, "transaction_jobs")
            if value:
                value["transaction_my_job"] = transaction_my_job
            else:
                value = {"transaction_my_job": transaction_my_job}
            await record.metadata_set(session, key="transaction_jobs", value=value)

        tx_job_to_send = TransactionJobToSend(job=transaction_my_job)

        return tx_job_to_send

    async def set_transaction_their_job(
        self, tx_job_received: TransactionJobToSend, connection: ConnRecord
    ):
        """Set transaction_their_job.

        Args:
            tx_job_received: The transaction job that is received from the other agent
            connection: connection to set metadata on
        """

        try:
            async with self._profile.session() as session:
                value = await connection.metadata_get(session, "transaction_jobs")
                if value:
                    value["transaction_their_job"] = tx_job_received.job
                else:
                    value = {"transaction_their_job": tx_job_received.job}
                await connection.metadata_set(
                    session, key="transaction_jobs", value=value
                )
        except StorageNotFoundError as err:
            raise TransactionManagerError(err.roll_up) from err

    async def endorsed_txn_post_processing(
        self,
        transaction: TransactionRecord,
        ledger_response: Optional[dict] = None,
        connection_record: Optional[ConnRecord] = None,
    ):
        """Store record in wallet, and kick off any required post-processing.

        Args:
            transaction: The transaction from which the schema/cred_def
                         would be stored in wallet.
            ledger_response: The ledger response
            connection_record: The connection record
        """

        if isinstance(ledger_response, str):
            ledger_response = json.loads(ledger_response)

        ledger = self._profile.inject(BaseLedger)
        if not ledger:
            reason = "No ledger available"
            if not self._profile.context.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise TransactionManagerError(reason)

        # setup meta_data to pass to future events, if necessary
        meta_data = transaction.meta_data
        meta_data["endorser"] = {
            "connection_id": transaction.connection_id,
        }

        is_anoncreds = self._profile.settings.get("wallet.type") == "askar-anoncreds"

        # write the wallet non-secrets record
        if ledger_response["result"]["txn"]["type"] == "101":
            # schema transaction
            schema_id = ledger_response["result"]["txnMetadata"]["txnId"]
            public_did = ledger_response["result"]["txn"]["metadata"]["from"]
            meta_data["context"]["schema_id"] = schema_id
            meta_data["context"]["public_did"] = public_did

            # Notify schema ledger write event
            if is_anoncreds:
                await AnonCredsIssuer(self._profile).finish_schema(
                    meta_data["context"]["job_id"],
                    meta_data["context"]["schema_id"],
                )
            else:
                await notify_schema_event(self._profile, schema_id, meta_data)

        elif ledger_response["result"]["txn"]["type"] == "102":
            # cred def transaction
            async with ledger:
                try:
                    schema_seq_no = str(ledger_response["result"]["txn"]["data"]["ref"])
                    schema_response = await shield(ledger.get_schema(schema_seq_no))
                except (IndyIssuerError, LedgerError) as err:
                    raise TransactionManagerError(err.roll_up) from err

            schema_id = schema_response["id"]
            cred_def_id = ledger_response["result"]["txnMetadata"]["txnId"]
            issuer_did = ledger_response["result"]["txn"]["metadata"]["from"]
            meta_data["context"]["schema_id"] = schema_id
            meta_data["context"]["cred_def_id"] = cred_def_id
            meta_data["context"]["issuer_did"] = issuer_did

            # Notify event
            if is_anoncreds:
                await AnonCredsIssuer(self._profile).finish_cred_def(
                    meta_data["context"]["job_id"],
                    meta_data["context"]["cred_def_id"],
                    meta_data["context"]["options"],
                )
            else:
                await notify_cred_def_event(self._profile, cred_def_id, meta_data)

        elif ledger_response["result"]["txn"]["type"] == "113":
            # revocation registry transaction
            rev_reg_id = ledger_response["result"]["txnMetadata"]["txnId"]
            meta_data["context"]["rev_reg_id"] = rev_reg_id
            if is_anoncreds:
                await AnonCredsRevocation(
                    self._profile
                ).finish_revocation_registry_definition(
                    meta_data["context"]["job_id"],
                    meta_data["context"]["rev_reg_id"],
                    meta_data["context"]["options"],
                )
            else:
                await notify_revocation_reg_endorsed_event(
                    self._profile, rev_reg_id, meta_data
                )

        elif ledger_response["result"]["txn"]["type"] == "114":
            # revocation entry transaction
            rev_reg_id = ledger_response["result"]["txn"]["data"]["revocRegDefId"]
            revoked = ledger_response["result"]["txn"]["data"]["value"].get("revoked", [])
            meta_data["context"]["rev_reg_id"] = rev_reg_id
            if is_anoncreds:
                await AnonCredsRevocation(self._profile).finish_revocation_list(
                    meta_data["context"]["job_id"], rev_reg_id, revoked
                )
            else:
                await notify_revocation_entry_endorsed_event(
                    self._profile, rev_reg_id, meta_data, revoked
                )

        elif ledger_response["result"]["txn"]["type"] == "1":
            # write DID to ledger
            did = ledger_response["result"]["txn"]["data"]["dest"]
            await notify_endorse_did_event(self._profile, did, meta_data)

        elif ledger_response["result"]["txn"]["type"] == "100":
            # write DID ATTRIB to ledger
            did = ledger_response["result"]["txn"]["data"]["dest"]
            await notify_endorse_did_attrib_event(self._profile, did, meta_data)

        else:
            # TODO unknown ledger transaction type, just ignore for now ...
            pass



import pytest
import pytest_asyncio

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...handlers.ping_response_handler import PingResponseHandler
from ...messages.ping_response import PingResponse


@pytest_asyncio.fixture
async def request_context():
    yield RequestContext.test_context(await create_test_profile())


class TestPingResponseHandler:
    @pytest.mark.asyncio
    async def test_ping_response(self, request_context):
        request_context.message_receipt = MessageReceipt()
        request_context.message = PingResponse()
        request_context.settings["debug.monitor_ping"] = True
        request_context.connection_ready = True
        handler = PingResponseHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)
        messages = responder.messages
        assert len(messages) == 0



"""Test Event Bus."""

import re
from unittest import mock

import pytest

from .. import event_bus as test_module
from ..event_bus import Event, EventBus

# pylint: disable=redefined-outer-name


@pytest.fixture
def event_bus():
    yield EventBus()


@pytest.fixture
def profile():
    yield mock.MagicMock()


@pytest.fixture
def event():
    event = Event(topic="anything", payload="payload")
    yield event


class MockProcessor:
    def __init__(self):
        self.profile = None
        self.event = None

    async def __call__(self, profile, event):
        self.profile = profile
        self.event = event


@pytest.fixture
def processor():
    yield MockProcessor()


def test_event(event):
    assert event.topic == "anything"
    assert event.payload == "payload"
    other = Event("anything", "payload")
    another = Event("nothing", "payload")
    and_another = Event("anything")
    assert event == other
    assert event != another
    assert event != and_another
    assert event != "random string"
    assert repr(event)


def test_sub_unsub(event_bus: EventBus, processor):
    """Test subscribe and unsubscribe."""
    event_bus.subscribe(re.compile(".*"), processor)
    assert event_bus.topic_patterns_to_subscribers
    assert event_bus.topic_patterns_to_subscribers[re.compile(".*")] == [processor]
    event_bus.unsubscribe(re.compile(".*"), processor)
    assert not event_bus.topic_patterns_to_subscribers


def test_unsub_idempotency(event_bus: EventBus, processor):
    """Test unsubscribe idempotency."""
    event_bus.subscribe(re.compile(".*"), processor)
    event_bus.unsubscribe(re.compile(".*"), processor)
    assert not event_bus.topic_patterns_to_subscribers
    event_bus.unsubscribe(re.compile(".*"), processor)
    assert not event_bus.topic_patterns_to_subscribers


def test_unsub_unsubbed_processor(event_bus: EventBus, processor):
    """Test unsubscribing an unsubscribed processor does not error."""
    event_bus.unsubscribe(re.compile(".*"), processor)
    event_bus.subscribe(re.compile(".*"), processor)
    another_processor = MockProcessor()
    event_bus.unsubscribe(re.compile(".*"), another_processor)


@pytest.mark.asyncio
async def test_sub_notify(event_bus: EventBus, profile, event, processor):
    """Test subscriber receives event."""
    event_bus.subscribe(re.compile(".*"), processor)
    await event_bus.notify(profile, event)
    assert processor.profile == profile
    assert processor.event == event


@pytest.mark.asyncio
async def test_sub_notify_error_logged_and_exec_continues(
    event_bus: EventBus,
    profile,
    event,
):
    """Test subscriber errors are logged but do not halt execution."""

    def _raise_exception(profile, event):
        raise Exception()

    processor = MockProcessor()
    bad_processor = _raise_exception
    event_bus.subscribe(re.compile(".*"), bad_processor)
    event_bus.subscribe(re.compile(".*"), processor)
    with mock.patch.object(
        test_module.LOGGER, "exception", mock.MagicMock()
    ) as mock_log_exc:
        await event_bus.notify(profile, event)

    mock_log_exc.assert_called_once_with("Error occurred while processing event")
    assert processor.profile == profile
    assert processor.event == event


@pytest.mark.parametrize(
    "pattern, topic",
    [
        ("test", "test"),
        (".*", "test"),
        ("topic::with::namespace", "topic::with::namespace::like::pieces"),
    ],
)
@pytest.mark.asyncio
async def test_sub_notify_regex_filtering(
    event_bus: EventBus, profile, processor, pattern, topic
):
    """Test events are filtered correctly."""
    event = Event(topic)
    event_bus.subscribe(re.compile(pattern), processor)
    await event_bus.notify(profile, event)
    assert processor.profile == profile
    assert processor.event == event


@pytest.mark.asyncio
async def test_sub_notify_no_match(event_bus: EventBus, profile, event, processor):
    """Test event not given to processor when pattern doesn't match."""
    event_bus.subscribe(re.compile("^$"), processor)
    await event_bus.notify(profile, event)
    assert processor.profile is None
    assert processor.event is None


@pytest.mark.asyncio
async def test_sub_notify_only_one(event_bus: EventBus, profile, event, processor):
    """Test only one subscriber is called when pattern matches only one."""
    processor1 = MockProcessor()
    event_bus.subscribe(re.compile(".*"), processor)
    event_bus.subscribe(re.compile("^$"), processor1)
    await event_bus.notify(profile, event)
    assert processor.profile == profile
    assert processor.event == event
    assert processor1.profile is None
    assert processor1.event is None


@pytest.mark.asyncio
async def test_sub_notify_both(event_bus: EventBus, profile, event, processor):
    """Test both subscribers are called when pattern matches both."""
    processor1 = MockProcessor()
    event_bus.subscribe(re.compile(".*"), processor)
    event_bus.subscribe(re.compile("anything"), processor1)
    await event_bus.notify(profile, event)
    assert processor.profile == profile
    assert processor.event == event
    assert processor1.profile == profile
    assert processor1.event == event


@pytest.mark.asyncio
async def test_wait_for_event_multiple_do_not_collide(event_bus: EventBus, profile):
    """Test multiple wait_for_event calls don't collide."""
    pattern = re.compile(".*")
    with event_bus.wait_for_event(profile, pattern) as event1:
        with event_bus.wait_for_event(profile, pattern) as event2:
            assert len(event_bus.topic_patterns_to_subscribers) == 1
            assert len(event_bus.topic_patterns_to_subscribers[pattern]) == 2
            event_bus.unsubscribe(
                pattern, event_bus.topic_patterns_to_subscribers[pattern][0]
            )
            assert len(event_bus.topic_patterns_to_subscribers[pattern]) == 1


@pytest.mark.asyncio
async def test_wait_for_event(event_bus: EventBus, profile, event):
    with event_bus.wait_for_event(profile, re.compile(".*")) as returned_event:
        await event_bus.notify(profile, event)
        assert await returned_event == event


@pytest.mark.asyncio
async def test_wait_for_event_condition(event_bus: EventBus, profile, event):
    with event_bus.wait_for_event(
        profile, re.compile(".*"), lambda e: e.payload == "asdf"
    ) as returned_event:
        # This shouldn't trigger our condition because payload == "payload"
        await event_bus.notify(profile, event)
        assert not returned_event.done()

        # This should trigger
        event = Event("asdF", "asdf")
        await event_bus.notify(profile, event)
        assert returned_event.done()
        assert await returned_event == event



TEST_DID_SOV = "did:sov:LjgpST2rjsoxYegQDRm7EL"
TEST_DID_KEY_ED25519 = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_DID_KEY_BLS_G2 = "did:key:zUC74E9UD2W6Q1MgPexCEdpstiCsY1Vbnyqepygk7McZVce38L1tGX7qZ2SgY4Zz2m9FUB4Xb5cEHSujks9XeKDzqe4QzW3CyyJ1cv8iBLNqU61EfkBoW2yEkg6VgqHTDtANYRS"

LD_PROOF_VC_DETAIL = {
    "credential": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "credentialSubject": {"test": "key"},
        "issuanceDate": "2021-04-12",
        "issuer": TEST_DID_KEY_ED25519,
    },
    "options": {
        "proofType": "Ed25519Signature2018",
        "created": "2019-12-11T03:50:55",
    },
}
LD_PROOF_VC_DETAIL_BBS = {
    "credential": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "credentialSubject": {"test": "key"},
        "issuanceDate": "2021-04-12",
        "issuer": TEST_DID_KEY_BLS_G2,
    },
    "options": {
        "proofType": "BbsBlsSignature2020",
        "created": "2019-12-11T03:50:55",
    },
}
LD_PROOF_VC_DETAIL_ED25519_2020 = {
    "credential": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "credentialSubject": {"test": "key"},
        "issuanceDate": "2021-04-12",
        "issuer": TEST_DID_KEY_ED25519,
    },
    "options": {
        "proofType": "Ed25519Signature2020",
        "created": "2019-12-11T03:50:55",
    },
}
LD_PROOF_VC = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
    ],
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "credentialSubject": {"test": "key"},
    "issuanceDate": "2021-04-12",
    "issuer": TEST_DID_KEY_ED25519,
    "proof": {
        "proofPurpose": "assertionMethod",
        "created": "2019-12-11T03:50:55",
        "type": "Ed25519Signature2018",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..Q6amIrxGiSbM7Ce6DxlfwLCjVcYyclas8fMxaecspXFUcFW9DAAxKzgHx93FWktnlZjM_biitkMgZdStgvivAQ",
    },
}
